# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import ast
from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Callable, Optional

from xcube.core.varexpr.error import VarExprError

Names = Mapping[str, Any]
UnaryFunction = Callable[[Any], Any]
BinaryFunction = Callable[[Any, Any], Any]

_UNARY_OPS: Mapping[str, UnaryFunction] = {
    "UAdd": lambda x: +x,
    "USub": lambda x: -x,
    "Invert": lambda x: ~x,
    "Not": lambda x: not x,
}

_BINARY_OPS: Mapping[str, BinaryFunction] = {
    "Add": lambda x, y: x + y,
    "Sub": lambda x, y: x - y,
    "Mult": lambda x, y: x * y,
    "Div": lambda x, y: x / y,
    "FloorDiv": lambda x, y: x // y,
    "Mod": lambda x, y: x % y,
    "Pow": lambda x, y: x**y,
    "LShift": lambda x, y: x << y,
    "RShift": lambda x, y: x >> y,
    "BitAnd": lambda x, y: x & y,
    "BitXor": lambda x, y: x ^ y,
    "BitOr": lambda x, y: x | y,
}

_COMPARISON_OPS: Mapping[str, BinaryFunction] = {
    "Eq": lambda x, y: x == y,
    "NotEq": lambda x, y: x != y,
    "Lt": lambda x, y: x < y,
    "LtE": lambda x, y: x <= y,
    "Gt": lambda x, y: x > y,
    "GtE": lambda x, y: x >= y,
    "In": lambda x, y: x in y,
    "NotIn": lambda x, y: x not in y,
    "Is": lambda x, y: x is y,
    "IsNot": lambda x, y: x is not y,
}


def parse(code: str) -> "VarExpr":
    # noinspection PyTypeChecker
    expr_node: ast.expr = ast.parse(code, mode="eval")
    return VarExprFactory().visit(expr_node)


def evaluate(code: str, names: Mapping[str, Any]) -> Any:
    var_expr = parse(code)
    return var_expr.evaluate(names)


class VarExpr:
    """Represents a node of a variable expression."""

    @abstractmethod
    def evaluate(self, names: Names) -> Any:
        """Evaluate the expression."""


class Constant(VarExpr):
    # noinspection PyShadowingBuiltins
    def __init__(self, value: Any):
        self.value = value

    def evaluate(self, names: Names) -> Any:
        return self.value


class Name(VarExpr):
    # noinspection PyShadowingBuiltins
    def __init__(self, id: str):
        self.id = id

    def evaluate(self, names: Names) -> Any:
        try:
            return names[self.id]
        except KeyError:
            raise VarExprError(f"name {self.id!r} is not defined")


class Attribute(VarExpr):
    # noinspection PyShadowingBuiltins
    def __init__(self, value: VarExpr, attr: str):
        self.value = value
        self.attr = attr

    def evaluate(self, names: Names) -> Any:
        if self.attr.startswith("_"):
            raise VarExprError(f"illegal use of protected attribute {self.attr!r}")
        value = self.value.evaluate(names)
        try:
            return getattr(value, self.attr)
        except AttributeError as e:
            raise VarExprError(f"{e}")


# noinspection PyShadowingBuiltins
class Subscript(VarExpr):
    def __init__(
        self,
        value: VarExpr,
        slice: VarExpr,
    ):
        self.value = value
        self.slice = slice

    def evaluate(self, names: Names) -> Any:
        value = self.value.evaluate(names)
        slice = self.slice.evaluate(names)
        try:
            return value[slice]
        except IndexError as e:
            raise VarExprError(f"{e}")


class Slice(VarExpr):
    def __init__(
        self,
        lower: Optional[VarExpr],
        upper: Optional[VarExpr],
        step: Optional[VarExpr],
    ):
        self.lower = lower
        self.upper = upper
        self.step = step

    def evaluate(self, names: Names) -> slice:
        lower = None if self.lower is None else self.lower.evaluate(names)
        upper = None if self.upper is None else self.upper.evaluate(names)
        step = None if self.step is None else self.step.evaluate(names)
        return slice(lower, upper, step)


class Call(VarExpr):
    def __init__(
        self, func: VarExpr, args: list[VarExpr], keywords: Mapping[str, VarExpr]
    ):
        self.func = func
        self.args = args
        self.keywords = keywords

    def evaluate(self, names: Names) -> Any:
        return self.func.evaluate(names)(
            *(arg.evaluate(names) for arg in self.args),
            **{value: arg.evaluate(names) for value, arg in self.keywords.items()},
        )


class UnaryOp(VarExpr):
    def __init__(self, op: UnaryFunction, operand: VarExpr):
        self.op = op
        self.operand = operand

    def evaluate(self, names: Names) -> Any:
        return self.op(self.operand.evaluate(names))


class BinOp(VarExpr):
    def __init__(self, left: VarExpr, op: BinaryFunction, right: VarExpr):
        self.left = left
        self.op = op
        self.right = right

    def evaluate(self, names: Names) -> Any:
        return self.op(self.left.evaluate(names), self.right.evaluate(names))


class Compare(VarExpr):
    def __init__(
        self, left: VarExpr, ops: list[BinaryFunction], comparators: list[VarExpr]
    ):
        self.left = left
        self.ops = ops
        self.comparators = comparators

    def evaluate(self, names: Names) -> bool:
        result = left = self.left.evaluate(names)
        for op, comparator in zip(self.ops, self.comparators):
            right = comparator.evaluate(names)
            result = op(left, right)
            if not result:
                break
            left = right
        return result


class BoolOp(VarExpr):
    def __init__(self, op: str, values: list[VarExpr]):
        self.op = op
        self.values = values

    def evaluate(self, names: Names) -> bool:
        if self.op == "Or":
            value = None
            for value in self.values:
                value = value.evaluate(names)
                if value:
                    break
            return value
        else:
            value = None
            for value in self.values:
                value = value.evaluate(names)
                if not value:
                    break
            return value


class IfExp(VarExpr):
    # noinspection SpellCheckingInspection
    def __init__(self, test: VarExpr, body: VarExpr, orelse: VarExpr):
        self.test = test
        self.body = body
        self.orelse = orelse

    def evaluate(self, names: Names) -> bool:
        return (
            self.body.evaluate(names)
            if self.test.evaluate(names)
            else self.orelse.evaluate(names)
        )


class Tuple(VarExpr):
    # noinspection SpellCheckingInspection
    def __init__(self, elts: list[VarExpr]):
        self.elts = elts

    def evaluate(self, names: Names) -> tuple:
        return tuple(elt.evaluate(names) for elt in self.elts)


class VarExprFactory(ast.NodeVisitor):
    def visit(self, node: ast.expr) -> VarExpr:
        return super().visit(node)

    def generic_visit(self, node: ast.AST):
        raise VarExprError(
            f"unsupported expression node of type {node.__class__.__name__!r}"
        )

    def visit_Expression(self, node: ast.Expression):
        # noinspection PyTypeChecker
        return super().visit(node.body)

    def visit_Constant(self, node: ast.Constant):
        return Constant(node.value)

    def visit_Name(self, node: ast.Name):
        return Name(node.id)

    def visit_Attribute(self, node: ast.Attribute):
        return Attribute(self.visit(node.value), node.attr)

    def visit_Subscript(self, node: ast.Subscript):
        return Subscript(self.visit(node.value), self.visit(node.slice))

    def visit_Slice(self, node: ast.Slice):
        return Slice(
            None if node.lower is None else self.visit(node.lower),
            None if node.upper is None else self.visit(node.upper),
            None if node.step is None else self.visit(node.step),
        )

    def visit_Call(self, node: ast.Call):
        return Call(
            self.visit(node.func),
            [self.visit(arg) for arg in node.args],
            {keyword.arg: self.visit(keyword.value) for keyword in node.keywords},
        )

    def visit_UnaryOp(self, node: ast.UnaryOp):
        return UnaryOp(_UNARY_OPS[node.op.__class__.__name__], self.visit(node.operand))

    def visit_BinOp(self, node: ast.BinOp):
        return BinOp(
            self.visit(node.left),
            _BINARY_OPS[node.op.__class__.__name__],
            self.visit(node.right),
        )

    def visit_Compare(self, node: ast.Compare):
        return Compare(
            self.visit(node.left),
            [_COMPARISON_OPS[op.__class__.__name__] for op in node.ops],
            [self.visit(c) for c in node.comparators],
        )

    def visit_BoolOp(self, node: ast.BoolOp):
        return BoolOp(node.op.__class__.__name__, [self.visit(v) for v in node.values])

    def visit_IfExp(self, node: ast.IfExp):
        return IfExp(
            self.visit(node.test), self.visit(node.body), self.visit(node.orelse)
        )

    def visit_Tuple(self, node: ast.Tuple):
        return Tuple([self.visit(elt) for elt in node.elts])



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
from collections.abc import Sequence

from xcube.server.server import Server
from xcube.util.config import load_configs


class ConfigChangeObserver:
    """An observer for configuration changes.

    Args:
        server: The server
        config_paths: Configuration file paths.
        check_after: Time in seconds between two observations.
    """

    def __init__(self, server: Server, config_paths: Sequence[str], check_after: float):
        self._server = server
        self._config_paths = config_paths
        self._check_after = check_after
        self._last_stats = None

    def check(self):
        last_stats = self._last_stats
        next_stats = [os.stat(config_path) for config_path in self._config_paths]
        if self._change_detected(last_stats, next_stats):
            next_config = load_configs(*self._config_paths)
            self._server.update(next_config)
        self._last_stats = next_stats
        self._server.call_later(self._check_after, self.check)

    @staticmethod
    def _change_detected(last_stats, next_stats) -> bool:
        if last_stats:
            for s1, s2 in zip(next_stats, last_stats):
                if s1.st_mtime != s2.st_mtime:
                    return True
        return False



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import asyncio
import concurrent.futures
import functools
import logging
import traceback
import urllib.parse
from collections.abc import Awaitable, Mapping, Sequence
from typing import Any, Callable, Optional, Type, Union

import tornado.escape
import tornado.httputil
import tornado.ioloop
import tornado.web

from xcube.constants import LOG, LOG_LEVEL_DETAIL
from xcube.server.api import (
    JSON,
    ApiError,
    ApiHandler,
    ApiRequest,
    ApiResponse,
    ApiRoute,
    ApiStaticRoute,
    ArgT,
    Context,
    ReturnT,
)
from xcube.server.config import get_reverse_url_prefix, get_url_prefix
from xcube.server.framework import Framework
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import (
    JsonBooleanSchema,
    JsonIntegerSchema,
    JsonObjectSchema,
    JsonStringSchema,
)
from xcube.version import version

SERVER_CTX_ATTR_NAME = "__xcube_server_ctx"


class TornadoFramework(Framework):
    """
    The Tornado web server framework.
    """

    def __init__(
        self,
        application: Optional[tornado.web.Application] = None,
        io_loop: Optional[tornado.ioloop.IOLoop] = None,
        shared_io_loop: bool = False,
    ):
        self._application = application or tornado.web.Application()
        self._io_loop = io_loop
        self._shared_io_loop = io_loop is not None and shared_io_loop
        self.configure_logging()

    @property
    def config_schema(self) -> Optional[JsonObjectSchema]:
        """Returns the JSON Schema for the configuration of the
        ``tornado.httpserver.HTTPServer.initialize()``
        method.
        """
        return JsonObjectSchema(
            properties=dict(
                tornado=JsonObjectSchema(
                    # See kwargs of tornado.httpserver.HTTPServer.initialize()
                    properties=dict(
                        xheaders=JsonBooleanSchema(default=False),
                        protocol=JsonStringSchema(nullable=True),
                        no_keep_alive=JsonBooleanSchema(default=False),
                        decompress_request=JsonBooleanSchema(default=False),
                        ssl_options=JsonObjectSchema(additional_properties=True),
                        chunk_size=JsonIntegerSchema(
                            nullable=True, exclusive_minimum=0
                        ),
                        max_header_size=JsonIntegerSchema(
                            nullable=True, exclusive_minimum=0
                        ),
                        max_body_size=JsonIntegerSchema(
                            nullable=True, exclusive_minimum=0
                        ),
                    ),
                    additional_properties=True,
                )
            )
        )

    @property
    def application(self) -> tornado.web.Application:
        return self._application

    @property
    def io_loop(self) -> tornado.ioloop.IOLoop:
        return self._io_loop or tornado.ioloop.IOLoop.current()

    def add_static_routes(self, api_routes: Sequence[ApiStaticRoute], url_prefix: str):
        handlers = []
        for api_route in api_routes:
            base_url = f"{url_prefix}{api_route.path}"
            default_filename = api_route.default_filename
            handlers.append(
                (f"{base_url}", tornado.web.RedirectHandler, {"url": f"{base_url}/"})
            )
            handlers.append(
                (
                    f"{base_url}/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": api_route.dir_path, "default_filename": default_filename},
                )
            )
            LOG.log(
                LOG_LEVEL_DETAIL,
                f"Added static route"
                f" {api_route.path!r}"
                f" from API {api_route.api_name!r}",
            )
        if handlers:
            self.application.add_handlers(".*$", handlers)

    def add_routes(self, api_routes: Sequence[ApiRoute], url_prefix: str):
        handlers = []

        for api_route in api_routes:
            handlers.append(
                (
                    url_prefix
                    + self.path_to_pattern(api_route.path)
                    + ("/?" if api_route.slash else ""),
                    TornadoRequestHandler,
                    {"api_route": api_route},
                )
            )
            LOG.log(
                LOG_LEVEL_DETAIL,
                f"Added route {api_route.path!r} from API {api_route.api_name!r}",
            )

        if handlers:
            self.application.add_handlers(".*$", handlers)

    def update(self, ctx: Context):
        setattr(self.application, SERVER_CTX_ATTR_NAME, ctx)

    def start(self, ctx: Context):
        config = ctx.config

        port = config["port"]
        address = config["address"]
        url_prefix = get_url_prefix(config)
        tornado_settings = config.get("tornado", {})

        self.application.listen(port, address=address, **tornado_settings)

        address_ = "127.0.0.1" if address == "0.0.0.0" else address
        # TODO: get test URL template from configuration
        test_url = f"http://{address_}:{port}{url_prefix}/openapi.html"
        LOG.info(f"Service running, listening on {address}:{port}")
        LOG.info(f"Try {test_url}")
        LOG.info(f"Press CTRL+C to stop service")

        if not self._shared_io_loop:
            self.io_loop.start()

    def stop(self, ctx: Context):
        if not self._shared_io_loop:
            self.io_loop.stop()

    def call_later(
        self, delay: Union[int, float], callback: Callable, *args, **kwargs
    ) -> object:
        return self.io_loop.call_later(delay, callback, *args, **kwargs)

    def run_in_executor(
        self,
        executor: Optional[concurrent.futures.Executor],
        function: Callable[..., ReturnT],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[ReturnT]:
        return self.io_loop.run_in_executor(
            executor,
            functools.partial(function, **kwargs) if kwargs else function,
            *args,
        )

    @staticmethod
    def configure_logging():
        # Configure Tornado loggers to use root handlers, so we
        # have a single log level and format. Root handlers are
        # assumed to be already configured by xcube CLI.
        for logger_name in [
            "tornado",
            "tornado.access",
            "tornado.application",
            "tornado.general",
        ]:
            log = logging.getLogger(logger_name)
            # Remove Tornado's own handlers
            for h in list(log.handlers):
                log.removeHandler(h)
                h.close()
            # Add root handlers configured by xcube CLI
            for h in list(logging.root.handlers):
                log.addHandler(h)
            # Use common log level
            log.setLevel(logging.root.level)

    @staticmethod
    def path_to_pattern(path: str):
        """
        Convert a string *pattern* where any occurrences of ``{NAME}``
        are replaced by an equivalent regex expression which will
        assign matching character groups to NAME. Characters match until
        one of the RFC 2396 reserved characters is found or the end of
        the *pattern* is reached.

        Args:
            path: URL path

        Returns: equivalent regex pattern

        Raises:
            ValueError: if *pattern* is invalid
        """
        var_pattern = r"(?P<%s>[^\;\/\?\:\@\&\=\+\$\,]+)"
        rest_var_pattern = r"\/?(?P<%s>.*)"
        num_rest_vars = 0
        rest_var_seen = False
        reg_expr = ""
        pos = 0
        while True:
            pos1 = path.find("{", pos)
            if pos1 >= 0:
                pos2 = path.find("}", pos1 + 1)
                if pos2 <= pos1:
                    raise ValueError('missing closing "}" in "%s"' % path)
                arg = path[pos1 + 1 : pos2]
                if arg.startswith("*"):
                    rest_var_seen = True
                    name = arg[1:]
                    pattern = rest_var_pattern
                    if pos1 > 0 and path[pos1 - 1] == "/":
                        # Consume a trailing "/" because it is
                        # covered in pattern
                        pos1 -= 1
                    num_rest_vars += 1
                else:
                    rest_var_seen = False
                    name = arg
                    pattern = var_pattern
                if not name.isidentifier():
                    raise ValueError(
                        '"{name}" in path must be a valid identifier,'
                        ' but got "%s"' % arg
                    )
                reg_expr += path[pos:pos1] + (pattern % name)
                pos = pos2 + 1
            else:
                pos2 = path.find("}", pos)
                if pos2 >= pos:
                    raise ValueError('missing opening "{" in "%s"' % path)
                rest_of_path = path[pos:]
                if rest_var_seen and rest_of_path:
                    raise ValueError(
                        f'wildcard variable must be last in path, but path was "{path}"'
                    )
                reg_expr += rest_of_path
                rest_var_seen = False
                break
        if num_rest_vars > 1:
            raise ValueError(
                "only a single wildcard variable is allowed,"
                f' but found {num_rest_vars} in path "{path}"'
            )
        return reg_expr


# noinspection PyAbstractClass
class TornadoRequestHandler(tornado.web.RequestHandler):
    def __init__(
        self,
        application: tornado.web.Application,
        request: tornado.httputil.HTTPServerRequest,
        **kwargs: Any,
    ):
        super().__init__(application, request)
        server_ctx = getattr(application, SERVER_CTX_ATTR_NAME, None)
        assert isinstance(server_ctx, Context)
        api_route: ApiRoute = kwargs.pop("api_route")
        ctx: Context = server_ctx.get_api_ctx(api_route.api_name)
        self._api_handler: ApiHandler = api_route.handler_cls(
            ctx,
            TornadoApiRequest(
                request,
                get_url_prefix(server_ctx.config),
                get_reverse_url_prefix(server_ctx.config),
            ),
            TornadoApiResponse(self),
            **api_route.handler_kwargs,
        )

    def set_default_headers(self):
        self.set_header("Server", f"xcube-server/{version}")
        # TODO: get from config
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "GET,PUT,DELETE,OPTIONS")
        self.set_header(
            "Access-Control-Allow-Headers",
            "x-requested-with,access-control-allow-origin,authorization,content-type",
        )

    def write_error(self, status_code: int, **kwargs: Any):
        valid_json_types = str, int, float, bool, type(None)
        error_info = {
            k: v for k, v in kwargs.items() if isinstance(v, valid_json_types)
        }
        error_info.update(status_code=status_code)
        if "exc_info" in kwargs:
            exc_type, exc_val, exc_tb = kwargs["exc_info"]
            error_info.update(
                exception=traceback.format_exception(exc_type, exc_val, exc_tb),
                message=str(exc_val),
            )
            if isinstance(exc_val, tornado.web.HTTPError) and exc_val.reason:
                error_info.update(reason=exc_val.reason)
        self.finish({"error": error_info})

    async def head(self, *args, **kwargs):
        await self._call_method("head", *args, **kwargs)

    async def get(self, *args, **kwargs):
        await self._call_method("get", *args, **kwargs)

    async def post(self, *args, **kwargs):
        await self._call_method("post", *args, **kwargs)

    async def put(self, *args, **kwargs):
        await self._call_method("put", *args, **kwargs)

    async def delete(self, *args, **kwargs):
        await self._call_method("delete", *args, **kwargs)

    async def options(self, *args, **kwargs):
        await self._call_method("options", *args, **kwargs)

    async def _call_method(self, method_name: str, *args, **kwargs):
        method = getattr(self._api_handler, method_name)
        try:
            if asyncio.iscoroutinefunction(method):
                await method(*args, **kwargs)
            else:
                method(*args, **kwargs)
        except ApiError as e:
            raise tornado.web.HTTPError(e.status_code, log_message=e.message) from e


class TornadoApiRequest(ApiRequest):
    def __init__(
        self,
        request: tornado.httputil.HTTPServerRequest,
        url_prefix: str = "",
        reverse_url_prefix: str = "",
    ):
        self._request = request
        self._url_prefix = url_prefix
        self._reverse_url_prefix = reverse_url_prefix
        self._is_query_lower_case = False
        # print("full_url:", self._request.full_url())
        # print("protocol:", self._request.protocol)
        # print("host:", self._request.host)
        # print("uri:", self._request.uri)
        # print("path:", self._request.path)
        # print("query:", self._request.query)

    def make_query_lower_case(self):
        self._is_query_lower_case = True

    @functools.cached_property
    def query(self) -> Mapping[str, Sequence[str]]:
        mapping = urllib.parse.parse_qs(self._request.query)
        if self._is_query_lower_case:
            mapping = {k.lower(): v for k, v in mapping.items()}
        return mapping

    # noinspection PyShadowingBuiltins
    def get_query_args(
        self, name: str, type: Optional[type[ArgT]] = None
    ) -> Sequence[ArgT]:
        name_lc = name.lower() if self._is_query_lower_case else name
        if not self.query or name_lc not in self.query:
            return []
        values = self.query[name_lc]
        if type is not None and type is not str:
            assert_true(callable(type), "type must be callable")
            try:
                return [type(v) for v in values]
            except (ValueError, TypeError):
                raise ApiError.BadRequest(
                    f"Query parameter {name!r} must have type {type.__name__!r}."
                )
        return values

    def url_for_path(
        self, path: str, query: Optional[str] = None, reverse: bool = False
    ) -> str:
        """Get the URL for given *path* and *query*.
        If the *reverse* flag is set, the configuration parameter
        ``reverse_url_prefix``, if provided, is used to construct the URL,
        otherwise only ``url_prefix``, if provided, is used.
        """
        prefix = self._url_prefix
        if reverse:
            prefix = self._reverse_url_prefix or prefix
        uri = ""
        if path:
            uri = path if path.startswith("/") else "/" + path
        if query:
            uri += "?" + query
        if "://" in prefix:
            # Absolute prefix
            return f"{prefix}{uri}"
        else:
            # Relative prefix
            protocol = self._request.protocol
            host = self._request.host
            return f"{protocol}://{host}{prefix}{uri}"

    @property
    def url(self) -> str:
        return self._request.full_url()

    @property
    def headers(self) -> Mapping[str, str]:
        return self._request.headers

    @property
    def body(self) -> bytes:
        return self._request.body

    @property
    def json(self) -> JSON:
        try:
            return tornado.escape.json_decode(self.body)
        except ValueError as e:
            raise ApiError.BadRequest(f"Body does not contain valid JSON: {e}") from e


class TornadoApiResponse(ApiResponse):
    def __init__(self, handler: TornadoRequestHandler):
        self._handler = handler

    def set_header(self, name: str, value: str):
        self._handler.set_header(name, value)

    def set_status(self, status_code: int, reason: Optional[str] = None):
        self._handler.set_status(status_code, reason=reason)

    def write(self, data: Union[str, bytes, JSON], content_type: Optional[str] = None):
        if data is not None:
            self._handler.write(data)
        # https://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.write
        # "If the given chunk is a dictionary, we write it as JSON and set the
        # Content-Type of the response to be application/json. (if you want to
        # send JSON as a different Content-Type, call set_header after calling
        # write())."
        if content_type is not None:
            self._handler.set_header("Content-Type", content_type)

    def finish(
        self, data: Union[str, bytes, JSON] = None, content_type: Optional[str] = None
    ):
        self.write(data, content_type)
        return self._handler.finish()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Api

from .config import CONFIG_SCHEMA
from .context import MetaContext

api = Api(
    "meta",
    description="Server information and maintenance operations",
    config_schema=CONFIG_SCHEMA,
    required_apis=["auth"],
    create_ctx=MetaContext,
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import logging
import math

import pyproj

PLUGIN_ENTRY_POINT_GROUP_NAME = "xcube_plugins"
PLUGIN_MODULE_PREFIX = "xcube_"
PLUGIN_MODULE_NAME = "plugin"
PLUGIN_MODULE_FUNCTION_NAME = "init_plugin"
PLUGIN_LOAD_TIME_WARN_LIMIT = 100  # milliseconds
PLUGIN_INIT_TIME__WARN_LIMIT = 100  # milliseconds

#: The extension point identifier for input processor extensions
EXTENSION_POINT_INPUT_PROCESSORS = "xcube.core.gen.iproc"
#: The extension point identifier for dataset I/O extensions
EXTENSION_POINT_DATASET_IOS = "xcube.core.dsio"
#: The extension point identifier for CLI command extensions
EXTENSION_POINT_CLI_COMMANDS = "xcube.cli"
#: The extension point identifier for data stores
EXTENSION_POINT_DATA_STORES = "xcube.core.store"
#: The extension point identifier for data openers
EXTENSION_POINT_DATA_OPENERS = "xcube.core.store.opener"
#: The extension point identifier for data writers
EXTENSION_POINT_DATA_WRITERS = "xcube.core.store.writer"
#: The extension point identifier for server APIs
EXTENSION_POINT_SERVER_APIS = "xcube.server.api"
#: The extension point identifier for server frameworks such as Tornado
EXTENSION_POINT_SERVER_FRAMEWORKS = "xcube.server.framework"

GLOBAL_GEO_EXTENT = -180.0, -90.0, 180.0, 90.0

WGS84_ELLIPSOID_SEMI_MAJOR_AXIS = 6378137.0
EARTH_EQUATORIAL_PERIMETER = 2.0 * math.pi * WGS84_ELLIPSOID_SEMI_MAJOR_AXIS

CRS_WKT_EPSG_4326 = """
GEOGCS["WGS 84",
    DATUM["WGS_1984",
        SPHEROID["WGS 84",6378137,298.257223563,
            AUTHORITY["EPSG","7030"]],
        AUTHORITY["EPSG","6326"]],
    PRIMEM["Greenwich",0,
        AUTHORITY["EPSG","8901"]],
    UNIT["degree",0.01745329251994328,
        AUTHORITY["EPSG","9122"]],
    AUTHORITY["EPSG","4326"]]
"""

CRS84 = "OGC:CRS84"
CRS_CRS84 = pyproj.crs.CRS.from_string(CRS84)

FORMAT_NAME_ZARR = "zarr"
FORMAT_NAME_NETCDF4 = "netcdf4"
FORMAT_NAME_CSV = "csv"
FORMAT_NAME_MEM = "mem"
FORMAT_NAME_LEVELS = "levels"
FORMAT_NAME_SCRIPT = "script"
FORMAT_NAME_EXCEL = "excel"

# Note: this list must be kept in-sync with xcube.core.reproject:NAME_TO_GDAL_RESAMPLE_ALG
RESAMPLING_METHOD_NAMES = {
    # Up-sampling
    "Nearest",
    "Bilinear",
    "Cubic",
    "CubicSpline",
    "Lanczos",
    # Down-sampling
    "Average",
    "Min",
    "Max",
    "Median",
    "Mode",
    "Q1",
    "Q3",
}

LOG_LEVEL_OFF_NAME = "OFF"
LOG_LEVEL_OFF = logging.CRITICAL + 5

LOG_LEVEL_DETAIL_NAME = "DETAIL"
LOG_LEVEL_DETAIL = logging.DEBUG + 5

LOG_LEVEL_TRACE_NAME = "TRACE"
LOG_LEVEL_TRACE = logging.DEBUG - 5

logging.addLevelName(LOG_LEVEL_DETAIL, LOG_LEVEL_DETAIL_NAME)
logging.addLevelName(LOG_LEVEL_TRACE, LOG_LEVEL_TRACE_NAME)

LOG_LEVELS = [
    LOG_LEVEL_OFF_NAME,
    "CRITICAL",
    "ERROR",
    "WARNING",
    "INFO",
    LOG_LEVEL_DETAIL_NAME,
    "DEBUG",
    LOG_LEVEL_TRACE_NAME,
]

DEFAULT_GENERAL_LOG_LEVEL = LOG_LEVEL_OFF
DEFAULT_XCUBE_LOG_LEVEL = logging.WARNING

GENERAL_LOG_FORMAT = "[%(levelname).1s %(asctime)s %(name)s] %(message)s"
XCUBE_LOG_FORMAT = "%(levelname)s: %(message)s"

# xcube logger
LOG = logging.getLogger("xcube")
LOG.setLevel(DEFAULT_XCUBE_LOG_LEVEL)

DEFAULT_SERVER_FRAMEWORK = "tornado"
DEFAULT_SERVER_PORT = 8080
DEFAULT_SERVER_ADDRESS = "0.0.0.0"

DEFAULT_COLOR_MAP_NAME = "viridis"
DEFAULT_COLOR_MAP_VALUE_RANGE = (0.0, 1.0)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import ast
import warnings
from typing import Any, Dict


def compute_array_expr(
    expr: str,
    namespace: dict[str, Any] = None,
    errors: str = "raise",
    result_name: str = None,
):
    """Safely evaluate a Python expression and return the result.

    The namespace is expected to contain variable references
    which, when evaluated, return numpy-like data arrays.

    For example, the Python expression
    "B01 <= 0 and not min(B01) > 2" contains the variable reference
    "B01". The expression is first transpiled to the new expression
    "np.logical_and(B01 <= 0, np.logical_not(np.fmin(B01) > 2))"
    which is then being evaluated instead.

    Args:
        expr: A valid Python expression.
        namespace: A dictionary that represents the namespace for the
            evaluation.
        result_name: Name of the result (used for error messages only)
        errors: How to deal with errors raised when evaluating the
            expression. May be one of "raise" or "warn".

    Returns:
        The result computed from the evaluating the expression.
    """
    expr = transpile_expr(expr, warn=errors == "warn")
    return compute_expr(
        expr, namespace=namespace, errors=errors, result_name=result_name
    )


def compute_expr(
    expr: str,
    namespace: dict[str, Any] = None,
    errors: str = "raise",
    result_name: str = None,
):
    """Safely evaluate a Python expression and return the result.

    Args:
        expr: A valid Python expression.
        namespace: A dictionary that represents the namespace for the
            evaluation.
        result_name: Name of the result (used for error messages only)
        errors: How to deal with errors raised when evaluating the
            expression. May be one of "raise" or "warn".

    Returns:
        The result computed from the evaluating the expression.
    """
    try:
        result = eval(expr, namespace, None)
    except Exception as e:
        result = None
        if result_name:
            msg = f"failed computing {result_name} from expression {expr!r}: {e}"
        else:
            msg = f"failed computing expression {expr!r}: {e}"
        if errors == "raise":
            raise ValueError(msg) from e
        if errors == "warn":
            warnings.warn(msg)
    return result


def transpile_expr(expr: str, warn=False) -> str:
    """Transpile a Python expression into a numpy/xarray array expression.

    Args:
        expr: The Python expression
        warn: If warnings shall be emitted

    Returns: The numpy array expression
    """
    return _ExprTranspiler(expr, warn).transpile()


# noinspection PyMethodMayBeStatic
class _ExprTranspiler:
    """Transpile a Python expression into a numpy/xarray array expression.

    See https://greentreesnakes.readthedocs.io/en/latest/nodes.html#expressions
    """

    _KEYWORDS = {
        "in",
        "not in",
        "is",
        "is not",
        "and",
        "or",
        "not",
        "True",
        "False",
        "None",
    }

    _OP_INFOS = {
        ast.IfExp: ("if", 10, "L"),
        ast.Eq: ("==", 100, "R"),
        ast.NotEq: ("!=", 100, "R"),
        ast.Lt: ("<", 100, "R"),
        ast.LtE: ("<=", 100, "R"),
        ast.Gt: (">", 100, "R"),
        ast.GtE: (">=", 100, "R"),
        ast.Is: ("is", 100, "R"),
        ast.IsNot: ("is not", 100, "R"),
        ast.In: ("in", 100, "R"),
        ast.NotIn: ("not in", 100, "R"),
        ast.Or: ("or", 300, "L"),
        ast.And: ("and", 400, "L"),
        ast.Not: ("not", 500, None),
        ast.UAdd: ("+", 600, None),
        ast.USub: ("-", 600, None),
        ast.Add: ("+", 600, "E"),
        ast.Sub: ("-", 600, "L"),
        ast.Mult: ("*", 700, "E"),
        ast.Div: ("/", 700, "L"),
        ast.FloorDiv: ("//", 700, "L"),
        ast.Mod: ("%", 800, "L"),
        ast.Pow: ("**", 900, "L"),
    }

    @classmethod
    def get_op_info(cls, op: ast.AST):
        return cls._OP_INFOS.get(type(op), (None, None, None))

    def __init__(self, expr: str, warn: bool):
        self.expr = expr
        self.warn = warn

    def transpile(self) -> str:
        return self._transpile(ast.parse(self.expr))

    def _transpile(self, node: ast.AST) -> str:
        if isinstance(node, ast.Module):
            return self._transpile(node.body[0])
        if isinstance(node, ast.Expr):
            return self._transpile(node.value)
        if isinstance(node, ast.Name):
            return self.transform_name(node)
        if isinstance(node, ast.Constant):
            return self.transform_name_constant(node)
        if isinstance(node, ast.Constant):
            return self.transform_num(node)
        if isinstance(node, ast.Attribute):
            pat = self.transform_attribute(node.value, node.attr, node.ctx)
            x = self._transpile(node.value)
            return pat.format(x=x)
        if isinstance(node, ast.Call):
            pat = self.transform_call(node.func, node.args)
            xes = {
                "x%s" % i: self._transpile(node.args[i]) for i in range(len(node.args))
            }
            return pat.format(**xes)
        if isinstance(node, ast.UnaryOp):
            pat = self.transform_unary_op(node.op, node.operand)
            arg = self._transpile(node.operand)
            return pat.format(x=arg)
        if isinstance(node, ast.BinOp):
            pat = self.transform_bin_op(node.op, node.left, node.right)
            x = self._transpile(node.left)
            y = self._transpile(node.right)
            return pat.format(x=x, y=y)
        if isinstance(node, ast.IfExp):
            pat = self.transform_if_exp(node.test, node.body, node.orelse)
            x = self._transpile(node.test)
            y = self._transpile(node.body)
            z = self._transpile(node.orelse)
            return pat.format(x=x, y=y, z=z)
        if isinstance(node, ast.BoolOp):
            pat = self.transform_bool_op(node.op, node.values)
            xes = {
                "x%s" % i: self._transpile(node.values[i])
                for i in range(len(node.values))
            }
            return pat.format(**xes)
        if isinstance(node, ast.Compare):
            pat = self.transform_compare(node.left, node.ops, node.comparators)
            xes = {"x0": self._transpile(node.left)}
            xes.update(
                {
                    "x%s" % (i + 1): self._transpile(node.comparators[i])
                    for i in range(len(node.comparators))
                }
            )
            return pat.format(**xes)
        raise ValueError(
            'unrecognized expression node %s in "%s"'
            % (node.__class__.__name__, self.expr)
        )

    def transform_name(self, name: ast.Name):
        return name.id

    def transform_name_constant(self, node: ast.Constant):
        return repr(node.value)

    def transform_num(self, node: ast.Constant):
        return str(node.n)

    def transform_call(self, func: ast.Name, args):
        args = ", ".join(["{x%d}" % i for i in range(len(args))])
        return f"{self.transform_function_name(func)}({args})"

    def transform_function_name(self, func: ast.Name):
        if isinstance(func, ast.Attribute):
            return f"{func.value.id}.{func.attr}"
        else:
            if func.id in ("where",):
                return "xr.%s" % func.id
            else:
                name = dict(min="fmin", max="fmax").get(func.id, func.id)
                return "np.%s" % name

    # noinspection PyUnusedLocal
    def transform_attribute(self, value: ast.AST, attr: str, ctx):
        return "{x}.%s" % attr

    def transform_unary_op(self, op, operand):
        name, precedence, _ = self.get_op_info(op)

        x = "{x}"

        if name == "not":
            return "np.logical_not(%s)" % x

        right_op = getattr(operand, "op", None)
        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence or (
                other_precedence == precedence and other_assoc is not None
            ):
                x = "({x})"

        if name in self._KEYWORDS:
            return f"{name} {x}"
        else:
            return f"{name}{x}"

    def transform_bin_op(self, op, left, right):
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        x = "{x}"
        y = "{y}"

        if name == "**":
            return f"np.power({x}, {y})"

        left_op = getattr(left, "op", None)
        right_op = getattr(right, "op", None)

        if left_op:
            _, other_precedence, other_assoc = self.get_op_info(left_op)
            if other_precedence < precedence or (
                other_precedence == precedence
                and assoc == "R"
                and other_assoc is not None
            ):
                x = "({x})"

        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence or (
                other_precedence == precedence
                and assoc == "L"
                and other_assoc is not None
            ):
                y = "({y})"

        return f"{x} {name} {y}"

    # noinspection PyUnusedLocal
    def transform_if_exp(self, test, body, orelse):
        return "xr.where({y}, {x}, {z})"

    def transform_bool_op(self, op, values):
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        if name == "and" or name == "or":
            expr = None
            for i in range(1, len(values)):
                expr = "np.logical_%s(%s, {x%d})" % (
                    name,
                    "{x0}" if i == 1 else expr,
                    i,
                )
            return expr

        xes = []
        for i in range(len(values)):
            value = values[i]
            x = "{x%d}" % i
            other_op = getattr(value, "op", None)
            if other_op:
                _, other_precedence, other_assoc = self.get_op_info(other_op)
                if (
                    i == 0
                    and other_precedence < precedence
                    or i > 0
                    and other_precedence <= precedence
                ):
                    x = "(%s)" % x
            xes.append(x)

        return (" %s " % name).join(xes)

    # Compare(left, ops, comparators
    def transform_compare(self, left, ops, comparators):
        if len(ops) != 1:
            raise ValueError(
                'expression "%s" uses an n-ary comparison,'
                " but only binary are supported" % self.expr
            )

        right = comparators[0]
        op = ops[0]
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        x = "{x0}"
        y = "{x1}"

        if self._is_nan(left):
            nan_op = y
        elif self._is_nan(right):
            nan_op = x
        else:
            nan_op = None

        if nan_op:
            if self.warn:
                warnings.warn(
                    "Use of NaN as operand with comparison"
                    ' "%s" is ambiguous: "%s"' % (name, self.expr)
                )
            if name == "==":
                return "np.isnan(%s)" % nan_op
            if name == "!=":
                return "np.logical_not(np.isnan(%s))" % nan_op

        left_op = getattr(left, "op", None)
        if left_op:
            name, other_precedence, assoc = _ExprTranspiler.get_op_info(left_op)
            if other_precedence < precedence:
                x = "(%s)" % x

        right_op = getattr(right, "op", None)
        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence or (
                other_precedence == precedence
                and assoc == "L"
                and other_assoc is not None
            ):
                y = "(%s)" % y

        return f"{x} {name} {y}"

    def _is_nan(self, node):
        return isinstance(node, ast.Name) and node.id == "NaN"



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

import altair as alt
import numpy as np
import pandas as pd
import pyproj
import shapely
import shapely.geometry
import shapely.ops
from chartlets import Component, Input, Output, State
from chartlets.components import Box, Button, CircularProgress, Select, VegaChart

from xcube.constants import CRS_CRS84
from xcube.core.geom import mask_dataset_by_geometry, normalize_geometry
from xcube.core.gridmapping import GridMapping
from xcube.server.api import Context
from xcube.webapi.viewer.contrib import Panel, get_dataset

panel = Panel(__name__, title="2D Histogram (Demo)", icon="equalizer", position=3)


# Number of bins in x and y directions.
# This results in columns of 4096 items.
# Vega Altair's maximum is 5000.
NUM_BINS_MAX = 64


@panel.layout(State("@app", "selectedDatasetId"))
def render_panel(ctx: Context, dataset_id: str | None = None) -> Component:
    dataset = get_dataset(ctx, dataset_id)

    plot = VegaChart(id="plot", chart=None, style={"paddingTop": 6})

    var_names, var_name_1, var_name_2 = get_var_select_options(dataset)

    select_var_1 = Select(
        id="select_var_1", label="Variable 1", value=var_name_1, options=var_names
    )
    select_var_2 = Select(
        id="select_var_2", label="Variable 2", value=var_name_2, options=var_names
    )

    button = Button(id="button", text="Update", style={"maxWidth": 100})

    controls = Box(
        children=[select_var_1, select_var_2, button],
        style={
            "display": "flex",
            "flexDirection": "row",
            "alignItems": "center",
            "gap": 6,
        },
    )

    return Box(
        children=[
            "Create or select a region shape in the map, then select two "
            "variables from the dropdowns, and press 'Update' to create "
            "a 2D histogram plot.",
            controls,
            plot
        ],
        style={
            "display": "flex",
            "flexDirection": "column",
            "alignItems": "center",
            "width": "100%",
            "height": "100%",
            "gap": 6,
        },
    )


@panel.callback(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    State("@app", "selectedPlaceGeometry"),
    State("select_var_1"),
    State("select_var_2"),
    Input("button", "clicked"),
    Output("plot", "chart"),
)
def update_plot(
    ctx: Context,
    dataset_id: str | None = None,
    time_label: float | None = None,
    place_geometry: str | None = None,
    var_1_name: str | None = None,
    var_2_name: str | None = None,
    _clicked: bool | None = None,  # trigger, will always be True
) -> alt.Chart | None:
    dataset = get_dataset(ctx, dataset_id)
    if dataset is None or not place_geometry or not var_1_name or not var_2_name:
        # TODO: set error message in panel UI
        print("panel disabled")
        return None

    if "time" in dataset.coords:
        if time_label:
            dataset = dataset.sel(time=pd.Timestamp(time_label[0:-1]), method="nearest")
        else:
            dataset = dataset.isel(time=-1)

    grid_mapping = GridMapping.from_dataset(dataset)
    place_geometry = normalize_geometry(place_geometry)
    if place_geometry is not None and not grid_mapping.crs.is_geographic:
        project = pyproj.Transformer.from_crs(
            CRS_CRS84, grid_mapping.crs, always_xy=True
        ).transform
        place_geometry = shapely.ops.transform(project, place_geometry)

    if (
        place_geometry is None
        or place_geometry.is_empty
        or isinstance(place_geometry, shapely.geometry.Point)
    ):
        # TODO: set error message in panel UI
        print("2-D histogram only works for geometries with a non-zero extent.")
        return

    dataset = mask_dataset_by_geometry(dataset, place_geometry)
    if dataset is None:
        # TODO: set error message in panel UI
        print("dataset is None after masking, invalid geometry?")
        return None

    var_1_data: np.ndarray = dataset[var_1_name].values.ravel()
    var_2_data: np.ndarray = dataset[var_2_name].values.ravel()
    var_1_range = [np.nanmin(var_1_data), np.nanmax(var_1_data)]
    var_2_range = [np.nanmin(var_2_data), np.nanmax(var_2_data)]
    num_bins = min(NUM_BINS_MAX, var_1_data.size)
    hist2d, x_edges, y_edges = np.histogram2d(
        var_1_data,
        var_2_data,
        bins=num_bins,
        range=np.array([var_1_range, var_2_range]),
        density=True,
    )
    # x and y 2D arrays with bin centers
    x_centers = (x_edges[:-1] + x_edges[1:]) / 2
    y_centers = (y_edges[:-1] + y_edges[1:]) / 2
    x, y = np.meshgrid(x_centers, y_centers)
    # z = hist2d
    z = np.where(hist2d > 0.0, hist2d, np.nan).T
    source = pd.DataFrame(
        {var_1_name: x.ravel(), var_2_name: y.ravel(), "z": z.ravel()}
    )
    # TODO: use edges or center coordinates as tick labels.
    x_centers = x_edges[0:-1] + np.diff(x_edges) / 2
    y_centers = y_edges[0:-1] + np.diff(y_edges) / 2
    # TODO: limit number of ticks on axes to, e.g., 10.
    # TODO: allow chart to be adjusted to available container (<div>) size.

    # Get the tick values
    x_num_ticks = 8
    x_tick_values = np.linspace(min(x_centers), max(x_centers), x_num_ticks)
    x_tick_values = np.array(
        [min(x_centers, key=lambda x: abs(x - t)) for t in x_tick_values]
    )

    num_ticks = 8
    y_tick_values = np.linspace(min(y_centers), max(y_centers), num_ticks)
    y_tick_values = np.array(
        [min(y_centers, key=lambda y: abs(y - t)) for t in y_tick_values]
    )

    chart = (
        alt.Chart(source)
        .mark_rect()
        .encode(
            x=alt.X(
                f"{var_1_name}:O",
                # axis=alt.Axis(values=x_axis),
                axis=alt.Axis(
                    labelAngle=45,
                    values=x_tick_values,
                    labelOverlap="greedy",
                    labelPadding=5,
                    format=".3f",
                ),
                # scale=alt.Scale(bins=x_centers),
                scale=alt.Scale(nice=True),
            ),
            y=alt.Y(
                f"{var_2_name}:O",
                sort="descending",
                # scale=alt.Scale(bins=y_centers),
                # axis=alt.Axis(values=y_axis),
                scale=alt.Scale(nice=True),
                axis=alt.Axis(
                    values=y_tick_values,
                    labelOverlap="greedy",
                    labelPadding=5,
                    format=".3f",
                ),
            ),
            color=alt.Color("z:Q", scale=alt.Scale(scheme="viridis"), title="Density"),
            tooltip=[var_1_name, var_2_name, "z:Q"],
        )
    ).properties(width=300, height=300)

    return chart


@panel.callback(
    Input("@app", "selectedDatasetId"),
    Input("@app", "selectedPlaceGeometry"),
    Output("button", "disabled"),
)
def set_button_disablement(
    _ctx: Context,
    dataset_id: str | None = None,
    place_geometry: str | None = None,
) -> bool:
    return not dataset_id or not place_geometry


@panel.callback(
    Input("@app", "selectedDatasetId"),
    State("select_var_1", "value"),
    State("select_var_2", "value"),
    Output("select_var_1", "options"),
    Output("select_var_1", "value"),
    Output("select_var_2", "options"),
    Output("select_var_2", "value"),
)
def populate_selects(
    ctx: Context,
    dataset_id: str | None = None,
    var_name_1: str | None = None,
    var_name_2: str | None = None,
) -> tuple[list, str | None, list, str | None]:
    dataset = get_dataset(ctx, dataset_id)
    var_names, var_name_1, var_name_2 = get_var_select_options(
        dataset, var_name_1, var_name_2
    )
    return var_names, var_name_1, var_names, var_name_2


def get_var_select_options(
    dataset,
    var_name_1: str | None = None,
    var_name_2: str | None = None,
) -> tuple[list, str | None, str | None]:
    if dataset is not None:
        var_names = [
            var_name
            for var_name, var in dataset.data_vars.items()
            if len(var.dims) >= 1
        ]
    else:
        var_names = []

    if var_names:
        if not var_name_1 or var_name_1 not in var_names:
            var_name_1 = var_names[0]
        if not var_name_2 or var_name_2 not in var_names:
            var_name_2 = var_names[0]

    return var_names, var_name_1, var_name_2


# TODO: Doesn't work. We need to ensure that show_progress() returns
#   before update_plot()
# @panel.callback(
#     Input("button", "clicked"),
#     Output("button", ""),
# )
def show_progress(
    _ctx: Context,
    _clicked: bool | None = None,  # trigger, will always be True
) -> alt.Chart | None:
    return CircularProgress(id="button", size=28)



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

import numpy as np
import xarray as xr


def compute_variables(ds, factor_chl, factor_tsm):
    chl_tsm_sum = factor_chl * ds.conc_chl + factor_tsm * ds.conc_tsm
    chl_tsm_sum.attrs.update(
        dict(
            units="-",
            long_name="Weighted sum of CHL nd TSM concentrations",
            description="Nonsense variable, for demo purpose only",
        )
    )

    chl_category = _categorize_chl(ds.conc_chl)
    chl_category.attrs.update(
        color_value_min=0.0,
        color_value_max=2.0,
        color_bar_name="spring",
        units="-",
        long_name="Chlorophyll bloom risk",
        description="Chlorophyll bloom risk in three categories: "
        "0: CHL < 3, "
        "1: 3 <= CHL < 4, "
        "2: CHL >= 4 mg/m^3",
    )

    return xr.Dataset(dict(chl_tsm_sum=chl_tsm_sum, chl_category=chl_category))


def _categorize_chl(chl):
    return xr.where(
        chl >= 4.0, 2, xr.where(chl >= 3.0, 1, xr.where(chl >= 0.0, 0, np.nan))
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os.path
from collections.abc import Mapping
from typing import Any, Dict, List, Optional, Union

from xcube.util.assertions import assert_given, assert_instance
from xcube.util.jsonschema import (
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObjectSchema,
    JsonStringSchema,
)

from ...util.config import load_json_or_yaml_config
from .assertions import assert_valid_config
from .error import DataStoreError
from .store import DataStore, new_data_store


def get_data_store_instance(
    store_id: str,
    store_params: dict[str, Any] = None,
    store_pool: "DataStorePool" = None,
) -> "DataStoreInstance":
    """Get a data store instance for identifier *store_id*.

    If *store_id* is prefixed by a "@", it is an "instance identifier".
    In this case the store instance is retrieved from
    the expected *store_pool* argument. Otherwise a new store instance
    is created using optional *store_params*.

    Args:
        store_id: Store identifier, may be prefixed by a "@" to indicate
            a store instance identifier.
        store_params: Store parameters, only valid if *store_id* is not
            an instance identifier.
        store_pool: A pool of configured store instances used if
            *store_id* is an instance identifier.

    Returns:
        a ``DataStoreInstance`` object

    Raises:
        DataStoreError: if a configured store does not exist
    """
    if store_id.startswith("@"):
        store_instance_id = store_id[1:]
        if store_pool is None:
            raise ValueError(
                f"store_pool must be given,"
                f' with store_id ("{store_id}")'
                f" referring to a configured store"
            )
        if store_params:
            raise ValueError(
                f"store_params cannot be given,"
                f' with store_id ("{store_id}")'
                f" referring to a configured store"
            )
        return store_pool.get_store_instance(store_instance_id)
    return DataStoreInstance(DataStoreConfig(store_id, store_params))


DATA_STORE_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        store_id=JsonStringSchema(min_length=1),
        store_params=JsonObjectSchema(additional_properties=True),
        title=JsonStringSchema(min_length=1),
        description=JsonStringSchema(min_length=1),
        cost_params=JsonObjectSchema(
            properties=dict(
                input_pixels_per_punit=JsonIntegerSchema(minimum=1),
                output_pixels_per_punit=JsonIntegerSchema(minimum=1),
                input_punits_weight=JsonNumberSchema(
                    exclusive_minimum=0.0, default=1.0
                ),
                output_punits_weight=JsonNumberSchema(
                    exclusive_minimum=0.0, default=1.0
                ),
            ),
            additional_properties=False,
            required=["input_pixels_per_punit", "output_pixels_per_punit"],
        ),
    ),
    required=["store_id"],
)

DATA_STORE_POOL_SCHEMA = JsonObjectSchema(
    additional_properties=DATA_STORE_CONFIG_SCHEMA,
)


class DataStoreConfig:
    """The configuration of a data store.
    The class is used by :class:`DataStorePool` to instantiate
    stores in a deferred manner.

    Args:
        store_id: the data store identifier
        store_params: optional store parameters
        title: a human-readable title for the store instance
        description: a human-readable description of the store instance
        user_data: optional user-data
    """

    def __init__(
        self,
        store_id: str,
        store_params: Mapping[str, Any] = None,
        title: str = None,
        description: str = None,
        user_data: Any = None,
    ):
        assert_given(store_id, name="store_id")
        if store_params is not None:
            assert_instance(store_params, dict, name="store_params")
        self._store_id = store_id
        self._store_params = store_params
        self._title = title
        self._description = description
        self._user_data = user_data

    @property
    def store_id(self) -> Optional[str]:
        return self._store_id

    @property
    def store_params(self) -> Optional[dict[str, Any]]:
        return self._store_params

    @property
    def title(self) -> Optional[str]:
        return self._title

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def user_data(self) -> Optional[Any]:
        return self._user_data

    @classmethod
    def from_dict(cls, data_store_config: dict[str, Any]) -> "DataStoreConfig":
        assert_valid_config(
            data_store_config, name="data_store_config", schema=DATA_STORE_CONFIG_SCHEMA
        )
        return DataStoreConfig(
            data_store_config["store_id"],
            store_params=data_store_config.get("store_params"),
            title=data_store_config.get("title"),
            description=data_store_config.get("description"),
        )

    def to_dict(self) -> dict[str, Any]:
        data_store_config = dict(store_id=self._store_id)
        if self._store_params:
            data_store_config.update(store_params=self._store_params)
        if self._title:
            data_store_config.update(name=self._title)
        if self._description:
            data_store_config.update(description=self._description)
        return data_store_config


class DataStoreInstance:
    """Internal class used by DataStorePool to maintain
    store configurations + instances.
    """

    def __init__(self, store_config: DataStoreConfig):
        assert_given(store_config, name="store_config")
        assert_instance(store_config, DataStoreConfig, name="store_config")
        self._store_config = store_config
        self._store: Optional[DataStore] = None

    @property
    def store_config(self) -> DataStoreConfig:
        return self._store_config

    @property
    def store(self) -> DataStore:
        if self._store is None:
            self._store = new_data_store(
                self._store_config.store_id, **(self._store_config.store_params or {})
            )
        return self._store

    def close(self):
        store = self._store
        if store is not None and hasattr(store, "close") and callable(store.close):
            store.close()


DataStoreConfigDict = dict[str, DataStoreConfig]
DataStoreInstanceDict = dict[str, DataStoreInstance]

DataStorePoolLike = Union[str, dict[str, Any], "DataStorePool"]


class DataStorePool:
    """A pool of configured data store instances.

    Actual data store instantiation only takes place lazily.
    A pool is may be created using it :meth:`from_dict` (or :meth:`from_file`)
    which receives a (JSON) dictionary that maps store instance names to
    store configurations:::

        {
            "<store_instance_id>": {
                "store_id": "<store_id>",
                "store_params": {
                    "<param_name>": <param_value>,
                    ...
                },
                "title": "<optional_human_readable_title>",
                "description": "<optional_human_readable_description>",
            },
            ...
        }

    Args::
        store_configs: A dictionary that maps store instance
            identifiers to to store configurations.
    """

    def __init__(self, store_configs: DataStoreConfigDict = None):
        if store_configs is not None:
            assert_instance(store_configs, dict, name="stores_configs")
        else:
            store_configs = {}
        self._instances: DataStoreInstanceDict = {
            k: DataStoreInstance(v) for k, v in store_configs.items()
        }

    @property
    def is_empty(self) -> bool:
        return len(self._instances) == 0

    @property
    def store_instance_ids(self) -> list[str]:
        return sorted([k for k, v in self._instances.items()])

    @property
    def store_configs(self) -> list[DataStoreConfig]:
        return [v.store_config for k, v in self._instances.items()]

    def get_store_instance_id(
        self, store_config: DataStoreConfig, strict_check: bool = False
    ) -> Optional[str]:
        assert_instance(store_config, DataStoreConfig, "store_config")
        for store_instance_id, instance in self._instances.items():
            if strict_check:
                if instance.store_config == store_config:
                    return store_instance_id
            else:
                if (
                    instance.store_config.store_id == store_config.store_id
                    and instance.store_config.store_params == store_config.store_params
                ):
                    return store_instance_id
        return None

    def has_store_config(self, store_config: DataStoreConfig) -> bool:
        return self.get_store_instance_id(store_config) is not None

    def has_store_instance(self, store_instance_id: str) -> bool:
        assert_instance(store_instance_id, str, "store_instance_id")
        return store_instance_id in self._instances

    def add_store_config(self, store_instance_id: str, store_config: DataStoreConfig):
        assert_instance(store_instance_id, str, "store_instance_id")
        assert_instance(store_config, DataStoreConfig, "store_config")
        if store_instance_id in self._instances:
            self._instances[store_instance_id].close()
        self._instances[store_instance_id] = DataStoreInstance(store_config)

    def remove_store_config(self, store_instance_id: str):
        self._assert_valid_instance_id(store_instance_id)
        self._instances[store_instance_id].close()
        del self._instances[store_instance_id]

    def remove_all_store_configs(self):
        self._instances.clear()

    def get_store_config(self, store_instance_id: str) -> DataStoreConfig:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id].store_config

    def get_store(self, store_instance_id: str) -> DataStore:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id].store

    def get_store_instance(self, store_instance_id: str) -> DataStoreInstance:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id]

    def close_all_stores(self):
        for instance in self._instances.values():
            instance.close()

    @classmethod
    def normalize(cls, data_store_pool: DataStorePoolLike) -> "DataStorePool":
        """Normalize given *data_store_pool* to an instance of
        :class:`DataStorePool`.

        If *data_store_pool* is already a DataStorePool it is returned as is.
        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the request is read from file using ``DataStorePool.from_file()``.
        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``DataStorePool.from_dict()``.

        Args:
            data_store_pool The data store pool instance,
                or data stores configuration file path, or data store pool
                JSON object.

        Raises:
            TypeError: if *data_store_pool* is not
                a ``CubeGeneratorRequest``, ``str``, or ``dict``.
        """
        if isinstance(data_store_pool, DataStorePool):
            return data_store_pool
        if isinstance(data_store_pool, str):
            return DataStorePool.from_file(data_store_pool)
        if isinstance(data_store_pool, dict):
            return DataStorePool.from_dict(data_store_pool)
        raise TypeError(
            "data_store_pool must be a str, dict, or a DataStorePool instance"
        )

    @classmethod
    def from_file(cls, path: str) -> "DataStorePool":
        _, ext = os.path.splitext(path)
        store_configs = load_json_or_yaml_config(path)
        return cls.from_dict(store_configs or {})

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "DataStorePool":
        DATA_STORE_POOL_SCHEMA.validate_instance(d)
        return cls({k: DataStoreConfig.from_dict(v) for k, v in d.items()})

    def to_dict(self) -> dict[str, Any]:
        return {
            instance_id: instance.store_config.to_dict()
            for instance_id, instance in self._instances.items()
        }

    def _assert_valid_instance_id(self, store_instance_id: str):
        assert_instance(store_instance_id, str, name="store_instance_id")
        if store_instance_id not in self._instances:
            raise DataStoreError(
                f'Configured data store instance "{store_instance_id}" not found.'
            )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Dict, Union

from xcube.util.config import load_json_or_yaml_config
from xcube.util.jsonschema import JsonObject, JsonObjectSchema, JsonStringSchema

DEFAULT_ENDPOINT_URL = "https://xcube-gen.brockmann-consult.de/api/v2/"

ServiceConfigLike = Union[str, dict, "ServiceConfig"]


class ServiceConfig(JsonObject):
    def __init__(
        self,
        endpoint_url: str = None,
        client_id: str = None,
        client_secret: str = None,
        access_token: str = None,
    ):
        endpoint_url = endpoint_url or DEFAULT_ENDPOINT_URL
        if not endpoint_url.endswith("/"):
            endpoint_url += "/"
        self.endpoint_url = endpoint_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = access_token

    @classmethod
    def normalize(cls, service_config: ServiceConfigLike) -> "ServiceConfig":
        """Normalize given *service_config* to an instance of
        :class:`ServiceConfig`.

        If *service_config* is already a ServiceConfig it is returned as is.

        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the configuration is read from file
        using ``ServiceConfig.from_file()``.´The file content may include
        template variables that are interpolated by environment variables,
        e.g. ``"${XCUBE_GEN_CLIENT_SECRET}"``.

        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``ServiceConfig.from_dict()``.

        Args:
            service_config: The remote configuration,
                or configuration file path, or configuration JSON object.

        Raises:
            TypeError: if *service_config* is not a ``CubeGeneratorRequest``,
                ``str``, or ``dict``.
        """
        if isinstance(service_config, ServiceConfig):
            return service_config
        if isinstance(service_config, str):
            return ServiceConfig.from_file(service_config)
        if isinstance(service_config, dict):
            return ServiceConfig.from_dict(service_config)
        raise TypeError(
            "service_config must be a str, dict, or a ServiceConfig instance"
        )

    @classmethod
    def from_file(cls, service_config_file: str) -> "ServiceConfig":
        service_config = load_json_or_yaml_config(service_config_file)
        cls.get_schema().validate_instance(service_config)
        return ServiceConfig(**service_config)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                endpoint_url=JsonStringSchema(min_length=1),
                client_id=JsonStringSchema(min_length=1),
                client_secret=JsonStringSchema(min_length=1),
                access_token=JsonStringSchema(min_length=1),
            ),
            additional_properties=False,
            required=[],
            factory=cls,
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Callable, Dict, Optional, Type

import jsonschema
import xarray as xr

from xcube.core.byoa import CodeConfig
from xcube.core.gridmapping import GridMapping
from xcube.util.jsonschema import JsonObjectSchema

from ..config import CubeConfig
from ..error import CubeGeneratorError
from ..processor import (
    METHOD_NAME_DATASET_PROCESSOR,
    METHOD_NAME_PARAMS_SCHEMA_GETTER,
    DatasetProcessor,
)
from .transformer import CubeTransformer, TransformedCube


class CubeUserCodeExecutor(CubeTransformer):
    """Execute user code."""

    def __init__(self, code_config: CodeConfig):
        user_code_callable = code_config.get_callable()
        user_code_callable_params = code_config.callable_params or {}

        if isinstance(user_code_callable, type):
            user_code_callable = self._get_callable_from_class(
                user_code_callable, user_code_callable_params
            )

        self._callable = user_code_callable
        self._callable_params = user_code_callable_params

    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        return self._callable(cube, **self._callable_params), gm, cube_config

    @classmethod
    def _get_callable_from_class(
        cls,
        process_class: type[Any],
        process_params: dict[str, Any],
    ) -> Callable:
        process_object = process_class()
        process_params_schema = None
        if isinstance(process_object, DatasetProcessor):
            process_callable = process_object.process_dataset
            process_params_schema = process_object.get_process_params_schema()
        else:
            process_callable = cls._get_user_code_callable(
                process_object, METHOD_NAME_DATASET_PROCESSOR, require=True
            )
            process_params_schema_getter = cls._get_user_code_callable(
                process_object, METHOD_NAME_PARAMS_SCHEMA_GETTER, require=False
            )
            if process_params_schema_getter is not None:
                process_params_schema = process_params_schema_getter()
        if process_params_schema is not None:
            if not isinstance(process_params_schema, JsonObjectSchema):
                raise CubeGeneratorError(
                    f"Parameter schema returned by"
                    f" user code class {process_class!r}"
                    f" must be an instance of {JsonObjectSchema!r}",
                    status_code=400,
                )
            try:
                process_params_schema.validate_instance(process_params)
            except jsonschema.ValidationError as e:
                raise CubeGeneratorError(
                    f"Invalid processing parameters: {e}", status_code=400
                ) from e
        return process_callable

    @classmethod
    def _get_user_code_callable(
        cls, user_code_object: Any, method_name: str, require: bool = True
    ) -> Optional[Callable]:
        if not hasattr(user_code_object, method_name):
            if not require:
                return None
            raise CubeGeneratorError(
                f"Missing method {method_name!r}"
                f" in user code class {type(user_code_object)!r}",
                status_code=400,
            )
        user_code_callable = getattr(user_code_object, method_name)
        if not callable(user_code_callable):
            raise CubeGeneratorError(
                f"Attribute {METHOD_NAME_DATASET_PROCESSOR!r}"
                f" of user code class {type(user_code_object)!r}"
                f" must be callable",
                status_code=400,
            )
        return user_code_callable



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import sys

import click

from xcube.constants import LOG


# noinspection PyShadowingBuiltins
@click.command(name="verify")
@click.argument("cube")
def verify(cube):
    """Perform cube verification.

    \b
    The tool verifies that CUBE
    * defines the dimensions "time", "lat", "lon";
    * has corresponding "time", "lat", "lon" coordinate variables and that they
      are valid, e.g. 1-D, non-empty, using correct units;
    * has valid  bounds variables for "time", "lat", "lon" coordinate
      variables, if any;
    * has any data variables and that they are valid, e.g. min. 3-D, all have
      same dimensions, have at least dimensions "time", "lat", "lon".
    * spatial coordinates and their corresponding bounds (if exist) are equidistant
       and monotonically increasing or decreasing.

    If INPUT is a valid xcube dataset, the tool returns exit code 0.
    Otherwise a violation report is written to stdout and the tool returns exit code 3.
    """
    return _verify(input_path=cube, monitor=print)


def _verify(input_path: str = None, monitor=None):
    from xcube.core.dsio import open_dataset
    from xcube.core.verify import verify_cube

    LOG.info(f"Opening cube from {input_path!r}...")
    with open_dataset(input_path) as cube:
        report = verify_cube(cube)

    if not report:
        monitor("INPUT is a valid cube.")
        return

    monitor("INPUT is not a valid cube due to the following reasons:")
    monitor("- " + "\n- ".join(report))
    # According to http://tldp.org/LDP/abs/html/exitcodes.html, exit code 3 is not reserved
    sys.exit(3)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


from xcube.server.api import Context
from xcube.webapi.common.context import ResourcesContext

from ...datasets.context import DatasetsContext


class CoveragesContext(ResourcesContext):
    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._datasets_ctx = server_ctx.get_api_ctx("datasets", cls=DatasetsContext)

    @property
    def datasets_ctx(self) -> DatasetsContext:
        return self._datasets_ctx



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonObjectSchema, JsonStringSchema

DEFAULT_CATALOG_ID = "xcube-server"
DEFAULT_CATALOG_TITLE = "xcube Server"
DEFAULT_CATALOG_DESCRIPTION = "Catalog of datasets served by xcube."

# ID, name, and description for the unified collection containing a feature
# for each datacube
DEFAULT_COLLECTION_ID = "datacubes"
DEFAULT_COLLECTION_TITLE = "Data cubes"
DEFAULT_COLLECTION_DESCRIPTION = "a collection of xcube datasets"

# As well as the unified collection, there's an individual collection for
# each datacube, with the same name as that datacube, and containing a single
# feature representing that datacube. This is the name of that single feature.
DEFAULT_FEATURE_ID = "datacube"

# Prefix for STAC and OGC endpoints
PATH_PREFIX = "/ogc"

COLLECTION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=JsonStringSchema(default=DEFAULT_COLLECTION_ID),
        Title=JsonStringSchema(default=DEFAULT_COLLECTION_TITLE),
        Description=JsonStringSchema(default=DEFAULT_COLLECTION_DESCRIPTION),
    ),
    additional_properties=False,
    required=[],
)

STAC_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=JsonStringSchema(default=DEFAULT_CATALOG_ID),
        Title=JsonStringSchema(default=DEFAULT_CATALOG_TITLE),
        Description=JsonStringSchema(default=DEFAULT_CATALOG_DESCRIPTION),
        Collection=COLLECTION_SCHEMA,
    ),
    additional_properties=False,
    required=[],
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        STAC=STAC_SCHEMA,
    )
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


import xarray as xr


def get_h_dim(ds: xr.Dataset) -> str:
    return [d for d in list(map(str, ds.sizes)) if d[:3].lower() in {"x", "lon"}][0]


def get_v_dim(ds: xr.Dataset) -> str:
    return [d for d in list(map(str, ds.sizes)) if d[:3].lower() in {"y", "lat"}][0]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.store import DataStorePool
from xcube.util.assertions import assert_instance
from xcube.util.progress import observe_progress

from .. import CubeConfig
from ..generator import CubeGenerator
from ..progress import ApiProgressCallbackObserver
from ..request import CubeGeneratorRequest, CubeGeneratorRequestLike
from ..response import CubeGeneratorResult, CubeInfoResult, CubeReference
from .combiner import CubesCombiner
from .helpers import is_empty_cube
from .informant import CubeInformant
from .mdadjuster import CubeMetadataAdjuster
from .opener import CubeOpener
from .rechunker import CubeRechunker
from .resamplert import CubeResamplerT
from .resamplerxy import CubeResamplerXY
from .subsetter import CubeSubsetter
from .transformer import CubeIdentity, transform_cube
from .usercode import CubeUserCodeExecutor
from .writer import CubeWriter


class LocalCubeGenerator(CubeGenerator):
    """Generator tool for data cubes.

    Creates cube views from one or more cube stores, resamples them to a
    common grid, optionally performs some cube transformation, and writes
    the resulting cube to some target cube store.

    Args:
        store_pool: An optional pool of pre-configured data stores
            referenced from *gen_config* input/output configurations.
        verbosity: Level of verbosity, 0 means off.
        raise_on_error: Whether to raise a CubeGeneratorError exception
            on generator failures. If False, the default, the returned
            result will have the "status" field set to "error" while
            other fields such as "message", "traceback", "output"
            provide more failure details.
    """

    def __init__(
        self,
        store_pool: DataStorePool = None,
        raise_on_error: bool = False,
        verbosity: int = 0,
    ):
        super().__init__(raise_on_error=raise_on_error, verbosity=verbosity)
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, "store_pool")

        self._store_pool = store_pool if store_pool is not None else DataStorePool()
        self._generated_data_id: Optional[str] = None
        self._generated_cube: Optional[xr.Dataset] = None
        self._generated_gm: Optional[GridMapping] = None

    @property
    def generated_data_id(self) -> Optional[str]:
        """Return the data identifier of the recently generated data cube."""
        return self._generated_data_id

    @property
    def generated_cube(self) -> Optional[xr.Dataset]:
        """Return the recently generated data cube."""
        return self._generated_cube

    @property
    def generated_gm(self) -> Optional[GridMapping]:
        """Return the recently generated data cube."""
        return self._generated_gm

    def _generate_cube(self, request: CubeGeneratorRequestLike) -> CubeGeneratorResult:
        request = CubeGeneratorRequest.normalize(request)
        request = request.for_local()

        observer = None
        if request.callback_config:
            observer = ApiProgressCallbackObserver(request.callback_config)
            observer.activate()

        try:
            return self.__generate_cube(request)
        finally:
            if observer is not None:
                observer.deactivate()

    def __generate_cube(self, request: CubeGeneratorRequest) -> CubeGeneratorResult:
        cube_config = (
            request.cube_config if request.cube_config is not None else CubeConfig()
        )

        opener = CubeOpener(cube_config, store_pool=self._store_pool)

        subsetter = CubeSubsetter()
        resampler_xy = CubeResamplerXY()
        resampler_t = CubeResamplerT()
        combiner = CubesCombiner(cube_config)
        rechunker = CubeRechunker()

        code_config = request.code_config
        if code_config is not None:
            code_executor = CubeUserCodeExecutor(code_config)
            post_rechunker = CubeRechunker()
        else:
            code_executor = CubeIdentity()
            post_rechunker = CubeIdentity()

        md_adjuster = CubeMetadataAdjuster()

        cube_writer = CubeWriter(request.output_config, store_pool=self._store_pool)

        num_inputs = len(request.input_configs)
        # Estimated workload:
        opener_work = 10
        resampler_t_work = 1
        resampler_xy_work = 20
        subsetter_work = 1
        combiner_work = num_inputs
        rechunker_work = 1
        executor_work = 1
        post_rechunker_work = 1
        metadata_adjuster_work = 1
        writer_work = 100  # this is where dask processing takes place
        total_work = (
            (opener_work + subsetter_work + resampler_t_work + resampler_xy_work)
            * num_inputs
            + combiner_work
            + rechunker_work
            + executor_work
            + post_rechunker_work
            + metadata_adjuster_work
            + writer_work
        )

        t_cubes = []
        with observe_progress("Generating cube", total_work) as progress:
            for input_config in request.input_configs:
                progress.will_work(opener_work)
                t_cube = opener.open_cube(input_config)

                progress.will_work(subsetter_work)
                t_cube = transform_cube(t_cube, subsetter, "subsetting")

                progress.will_work(resampler_t_work)
                t_cube = transform_cube(t_cube, resampler_t, "resampling in time")

                progress.will_work(resampler_xy_work)
                t_cube = transform_cube(t_cube, resampler_xy, "resampling in space")

                t_cubes.append(t_cube)

            progress.will_work(combiner_work)
            t_cube = combiner.combine_cubes(t_cubes)

            progress.will_work(rechunker_work)
            t_cube = transform_cube(t_cube, rechunker, "rechunking")

            progress.will_work(executor_work)
            t_cube = transform_cube(t_cube, code_executor, "executing user code")

            progress.will_work(post_rechunker_work)
            t_cube = transform_cube(t_cube, post_rechunker, "post-rechunking")

            progress.will_work(metadata_adjuster_work)
            t_cube = transform_cube(t_cube, md_adjuster, "adjusting metadata")

            progress.will_work(writer_work)
            cube, gm, _ = t_cube
            if not is_empty_cube(cube):
                data_id, cube = cube_writer.write_cube(cube, gm)
                self._generated_data_id = data_id
                self._generated_cube = cube
                self._generated_gm = gm
            else:
                self._generated_data_id = None
                self._generated_cube = None
                self._generated_gm = None

        total_time = progress.state.total_time

        if self._generated_data_id is not None:
            return CubeGeneratorResult(
                status="ok",
                status_code=201,
                result=CubeReference(data_id=data_id),
                message=f"Cube generated successfully after {total_time:.2f} seconds",
            )
        else:
            return CubeGeneratorResult(
                status="warning",
                status_code=422,
                message=f"An empty cube has been generated"
                f" after {total_time:.2f} seconds."
                f" No data has been written at all.",
            )

    def _get_cube_info(self, request: CubeGeneratorRequestLike) -> CubeInfoResult:
        request = CubeGeneratorRequest.normalize(request)
        informant = CubeInformant(
            request=request.for_local(), store_pool=self._store_pool
        )
        cube_info = informant.generate()
        return CubeInfoResult(result=cube_info, status="ok", status_code=200)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiError, ApiHandler
from xcube.util.assertions import assert_true
from xcube.util.cmaps import DEFAULT_CMAP_NAME

from ..places import PATH_PARAM_PLACE_GROUP_ID
from .api import api
from .context import DatasetsContext
from .controllers import (
    find_dataset_places,
    get_color_bars,
    get_dataset,
    get_dataset_coordinates,
    get_dataset_place_group,
    get_datasets,
    get_legend,
)

PATH_PARAM_DATASET_ID = {
    "name": "datasetId",
    "in": "path",
    "description": "Dataset identifier",
    "schema": {"type": "string"},
}

PATH_PARAM_VAR_NAME = {
    "name": "varName",
    "in": "path",
    "description": "Variable name",
    "schema": {"type": "string"},
}

QUERY_PARAM_CRS = {
    "name": "crs",
    "in": "query",
    "description": "The tile grid's spatial CRS",
    "schema": {"type": "string", "enum": ["EPSG:3857", "CRS84"], "default": "CRS84"},
}

QUERY_PARAM_VMIN = {
    "name": "vmin",
    "in": "query",
    "description": "Minimum value of variable for color mapping",
    "schema": {"type": "number", "default": 0},
}

QUERY_PARAM_VMAX = {
    "name": "vmax",
    "in": "query",
    "description": "Maximum value of variable for color mapping",
    "schema": {"type": "number", "default": 1},
}

QUERY_PARAM_CMAP = {
    "name": "cmap",
    "in": "query",
    "description": "Name of the (matplotlib) color mapping",
    "schema": {"type": "string", "default": DEFAULT_CMAP_NAME},
}

QUERY_PARAM_NORM = {
    "name": "norm",
    "in": "query",
    "description": "Name of the data normalisation applied before color mapping",
    "schema": {"enum": ["lin", "log", "cat"], "default": "lin"},
}


@api.route("/datasets")
class DatasetsHandler(ApiHandler[DatasetsContext]):
    """List the published datasets."""

    @api.operation(
        operation_id="getDatasets",
        summary="Get all datasets.",
        parameters=[
            {
                "name": "details",
                "in": "query",
                "description": "Whether to load dataset details",
                "schema": {
                    "type": "string",
                    "enum": ["0", "1"],
                },
            },
            {
                "name": "point",
                "in": "query",
                "description": "Only get datasets intersecting"
                ' given point. Format is "lon,lat",'
                ' for example "11.2,52.3"',
                "schema": {
                    "type": "string",
                },
            },
        ],
    )
    def get(self):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        details = self.request.get_query_arg("details", default=False)
        point = self.request.get_query_arg("point", default=None)
        if isinstance(point, str):
            try:
                point = tuple(map(float, point.split(",")))
                assert_true(len(point) == 2, "must be pair of two floats")
            except (ValueError, TypeError) as e:
                raise ApiError.BadRequest(f"illegal point: {e}")
        response = get_datasets(
            self.ctx,
            details=details,
            point=point,
            base_url=self.request.reverse_base_url,
            granted_scopes=granted_scopes,
        )
        self.response.finish(response)


@api.route("/datasets/{datasetId}")
class DatasetHandler(ApiHandler[DatasetsContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getDataset",
        summary="Get the details of a dataset.",
        parameters=[PATH_PARAM_DATASET_ID],
    )
    async def get(self, datasetId: str):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        result = get_dataset(
            self.ctx,
            datasetId,
            base_url=self.request.reverse_base_url,
            granted_scopes=granted_scopes,
        )
        self.response.set_header("Content-Type", "application/json")
        await self.response.finish(result)


@api.route("/datasets/{datasetId}/coords/{dimName}")
class DatasetCoordsHandler(ApiHandler[DatasetsContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getDatasetCoordinates",
        summary="Get the coordinates for a dimension of a dataset.",
        parameters=[PATH_PARAM_DATASET_ID],
    )
    async def get(self, datasetId: str, dimName: str):
        result = get_dataset_coordinates(self.ctx, datasetId, dimName)
        self.response.set_header("Content-Type", "application/json")
        await self.response.finish(result)


# noinspection PyPep8Naming
@api.route("/datasets/{datasetId}/places/{placeGroupId}")
class DatasetPlaceGroupHandler(ApiHandler[DatasetsContext]):
    """Get places for given dataset and place group."""

    @api.operation(
        operation_id="getDatasetPlaceGroup",
        summary="Get places for given dataset and place group.",
        parameters=[PATH_PARAM_DATASET_ID, PATH_PARAM_PLACE_GROUP_ID],
    )
    def get(self, datasetId: str, placeGroupId: str):
        response = get_dataset_place_group(
            self.ctx, datasetId, placeGroupId, self.request.reverse_base_url
        )
        self.response.finish(response)


# noinspection PyPep8Naming
@api.route("/places/{placeGroupId}/{datasetId}")
class PlacesForDatasetHandler(ApiHandler[DatasetsContext]):
    @api.operation(
        operation_id="findPlacesForDataset",
        tags=["places"],
        summary="Find places in place group for bounding box of given dataset.",
        parameters=[PATH_PARAM_PLACE_GROUP_ID, PATH_PARAM_DATASET_ID],
    )
    def get(self, placeGroupId: str, datasetId: str):
        query_expr = self.request.get_query_arg("query", default=None)
        comb_op = self.request.get_query_arg("comb", default="and")
        response = find_dataset_places(
            self.ctx,
            placeGroupId,
            datasetId,
            self.request.reverse_base_url,
            query_expr=query_expr,
            comb_op=comb_op,
        )
        self.response.finish(response)


# noinspection PyPep8Naming
class LegendHandler(ApiHandler[DatasetsContext]):
    async def get(self, datasetId: str, varName: str):
        legend = await self.ctx.run_in_executor(
            None,
            get_legend,
            self.ctx,
            datasetId,
            varName,
            {k: v[0] for k, v in self.request.query.items()},
        )
        self.response.set_header("Content-Type", "image/png")
        await self.response.finish(legend)


LEGEND_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    {
        "name": "width",
        "in": "query",
        "description": "Width of the legend in pixels",
        "schema": {"type": "number", "default": 256},
    },
    {
        "name": "height",
        "in": "query",
        "description": "Height of the legend in pixels",
        "schema": {"type": "number", "default": 16},
    },
]


# noinspection PyPep8Naming
@api.route("/datasets/{datasetId}/vars/{varName}/legend.png")
class OldLegendHandler(LegendHandler):
    @api.operation(
        operation_id="getLegendForVariable",
        summary="Get the legend as PNG used for the tiles"
        " for given variable. Deprecated!",
        parameters=LEGEND_PARAMETERS,
    )
    async def get(self, datasetId: str, varName: str):
        await super().get(datasetId, varName)


# noinspection PyPep8Naming
@api.route("/tiles/{datasetId}/{varName}/legend")
class NewLegendHandler(LegendHandler):
    @api.operation(
        operation_id="getLegendForTiles",
        tags=["tiles"],
        summary="Get the legend as PNG used for the tiles for given variable.",
        parameters=LEGEND_PARAMETERS,
    )
    async def get(self, datasetId: str, varName: str):
        await super().get(datasetId, varName)


# TODO (forman): move as endpoint "styles/colorbars" into API "styles"


@api.route("/colorbars")
class StylesColorBarsHandler(ApiHandler):
    """Get available color bars."""

    @api.operation(
        operation_id="getColorBars",
        summary="Get available color bars.",
        tags=["styles"],
    )
    def get(self):
        response = get_color_bars(self.ctx, "application/json")
        self.response.finish(response)


@api.route("/colorbars.html")
class StylesColorBarsHtmlHandler(ApiHandler):
    """Show available color bars."""

    @api.operation(
        operation_id="showColorBars",
        summary="Show available color bars.",
        tags=["styles"],
    )
    def get(self):
        response = get_color_bars(self.ctx, "text/html")
        self.response.finish(response)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import datetime

from xcube.server.api import ApiContext, Context
from xcube.server.server import ServerContext


class MetaContext(ApiContext):
    _start_time = None

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        # API contexts are instantiated every time the config changes
        if self._start_time is None:
            self._start_time = self.now()
        self._update_time = self.now()

    @property
    def server_ctx(self) -> ServerContext:
        server_ctx = super().server_ctx
        assert isinstance(server_ctx, ServerContext)
        return server_ctx

    @property
    def start_time(self) -> str:
        return self._start_time

    @property
    def update_time(self) -> str:
        return self._update_time

    @property
    def current_time(self) -> str:
        return self.now()

    @staticmethod
    def now():
        return datetime.datetime.now().isoformat()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .helpers import get_dataset, get_datasets_ctx
from .panel import Panel



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import inspect
from typing import Any, Callable, Dict

from xcube.server.api import ApiError

from .context import ComputeContext
from .op.info import OpInfo


def get_compute_operations(ctx: ComputeContext):
    ops = ctx.op_registry.ops
    return {"operations": [encode_op(op_id, f) for op_id, f in ops.items()]}


def get_compute_operation(ctx: ComputeContext, op_id: str):
    op = ctx.op_registry.get_op(op_id)
    if op is None:
        raise ApiError.NotFound(f"operation {op_id!r} not found")
    return encode_op(op_id, op)


def encode_op(op_id: str, op: Callable) -> dict[str, Any]:
    op_info = OpInfo.get_op_info(op)
    op_json = {"operationId": op_id, "parametersSchema": op_info.params_schema}
    doc = inspect.getdoc(op)
    if doc:
        op_json.update(description=doc)
    return op_json



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonArraySchema, JsonObjectSchema
from xcube.webapi.common.schemas import IDENTIFIER_SCHEMA, PATH_SCHEMA, STRING_SCHEMA

PLACE_GROUP_JOIN_SCHEMA = JsonObjectSchema(
    properties=dict(
        Property=IDENTIFIER_SCHEMA,
        Path=PATH_SCHEMA,
    ),
    required=[
        "Property",
        "Path",
    ],
    additional_properties=False,
)

PLACE_GROUP_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=IDENTIFIER_SCHEMA,
        Title=STRING_SCHEMA,
        Path=PATH_SCHEMA,
        Query=STRING_SCHEMA,
        Join=PLACE_GROUP_JOIN_SCHEMA,
        CharacterEncoding=STRING_SCHEMA,
        PropertyMapping=JsonObjectSchema(additional_properties=PATH_SCHEMA),
        PlaceGroupRef=IDENTIFIER_SCHEMA,
    ),
    required=[
        # Either we have
        #   'Identifier',
        #   'Path',
        # or we have
        #   'Identifier',
        #   'Query',
        # or we must specify
        #   'PlaceGroupRef',
    ],
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(PlaceGroups=JsonArraySchema(items=PLACE_GROUP_SCHEMA))
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import abc
import fnmatch
import importlib
import importlib.metadata
import pkgutil
import sys
import time
import traceback
import warnings
from typing import Any, Callable, Dict, List, Optional

from xcube.constants import (
    PLUGIN_ENTRY_POINT_GROUP_NAME,
    PLUGIN_INIT_TIME__WARN_LIMIT,
    PLUGIN_LOAD_TIME_WARN_LIMIT,
    PLUGIN_MODULE_FUNCTION_NAME,
    PLUGIN_MODULE_NAME,
    PLUGIN_MODULE_PREFIX,
)
from xcube.util.extension import Extension, ExtensionRegistry

_PLUGIN_REGISTRY_INIT = dict()

# Mapping of xcube entry point names
# to JSON-serializable plugin meta-information.
_PLUGIN_REGISTRY: dict[str, dict[str, Any]] = _PLUGIN_REGISTRY_INIT


def init_plugins() -> None:
    """Load plugins if not already done."""
    global _PLUGIN_REGISTRY
    if _PLUGIN_REGISTRY is _PLUGIN_REGISTRY_INIT:
        _PLUGIN_REGISTRY = load_plugins()


def get_plugins() -> dict[str, dict]:
    """Get mapping of "xcube_plugins" entry point names
    to JSON-serializable plugin meta-information.
    """
    init_plugins()
    global _PLUGIN_REGISTRY
    return dict(_PLUGIN_REGISTRY)


def get_extension_registry() -> ExtensionRegistry:
    """Get populated extension registry."""
    from xcube.util.extension import get_extension_registry

    init_plugins()
    return get_extension_registry()


def discover_plugin_modules(module_prefixes=None):
    module_prefixes = module_prefixes or [PLUGIN_MODULE_PREFIX]
    entry_points = []
    for module_finder, module_name, ispkg in pkgutil.iter_modules():
        if any(
            [module_name.startswith(module_prefix) for module_prefix in module_prefixes]
        ):
            # Note: Consider turning this into debug log,
            #  but logging is not yet configured at this point.
            # print(f'xcube plugin module found: {module_name}')
            entry_points.append(_ModuleEntryPoint(module_name))

    # Discover plugins if they are installed in a editable mode.
    # See https://github.com/xcube-dev/xcube/issues/963
    for module_prefix in module_prefixes:
        for path_hook in fnmatch.filter(
            sys.path, f"__editable__.{module_prefix}*.finder.__path_hook__"
        ):
            module_name = path_hook.split("-")[0].split(".")[1]
            entry_points.append(_ModuleEntryPoint(module_name))

    return entry_points


def load_plugins(
    entry_points=None, ext_registry: Optional[ExtensionRegistry] = None
) -> dict[str, dict[str, Any]]:
    if ext_registry is None:
        from xcube.util.extension import get_extension_registry

        ext_registry = get_extension_registry()

    plugins = {}

    if entry_points:
        _collect_plugins(entry_points, ext_registry, True, plugins)
    else:
        # verbose=True for specified xcube plugins.
        _collect_plugins(
            list(importlib.metadata.entry_points(group=PLUGIN_ENTRY_POINT_GROUP_NAME)),
            ext_registry,
            True,
            plugins,
        )
        # verbose=False for auto-detected xcube plugins,
        # where package name starts with "xcube_" but an
        # entry point is not explicitly specified.
        _collect_plugins(discover_plugin_modules(), ext_registry, False, plugins)

    return plugins


def _collect_plugins(
    entry_points: list[Any],
    ext_registry: ExtensionRegistry,
    verbose: bool,
    plugins: dict[str, dict[str, Any]],
):
    for entry_point in entry_points:
        # Note: Consider turning this into debug log,
        #  but logging is not yet configured at this point.
        #  print(f'loading xcube plugin {entry_point.name!r}')

        t0 = time.perf_counter()

        # noinspection PyBroadException
        try:
            plugin_init_function = entry_point.load()
        except Exception as e:
            if verbose:
                _emit_warning_for_error(entry_point, e)
            continue

        if plugin_init_function is None:
            # Not a plugin
            continue

        if verbose:
            millis = int(1000 * (time.perf_counter() - t0))
            if millis >= PLUGIN_LOAD_TIME_WARN_LIMIT:
                _emit_warning_for_slow_load(entry_point, millis)

        if not callable(plugin_init_function):
            if verbose:
                _emit_warning_on_init_function(entry_point, plugin_init_function)
            continue

        t0 = time.perf_counter()

        # noinspection PyBroadException
        try:
            plugin_init_function(ext_registry)
        except Exception as e:
            if verbose:
                _emit_warning_for_error(entry_point, e)
            continue

        if verbose:
            millis = int(1000 * (time.perf_counter() - t0))
            if millis >= PLUGIN_INIT_TIME__WARN_LIMIT:
                _emit_warning_for_slow_load(entry_point, millis)

        plugins[entry_point.name] = {
            "name": entry_point.name,
            "doc": plugin_init_function.__doc__,
        }


def _emit_warning_on_init_function(entry_point, plugin_init_function):
    # We use warning and not raise to allow loading
    # xcube despite a broken plugin. Raise would stop xcube.
    warnings.warn(
        f"xcube plugin {entry_point.name!r}"
        f" must be callable"
        f" but got a {type(plugin_init_function)!r}"
    )


def _emit_warning_for_slow_load(entry_point, millis):
    warnings.warn(
        f"Initializing xcube plugin {entry_point.name!r}"
        f" took {millis} ms,"
        f" consider code optimization."
        f" (For example, avoid eager import of packages,"
        f" consider lazy loading of resources, etc.)"
    )


def _emit_warning_for_error(entry_point, e):
    # We use warning and not raise to allow loading
    # xcube despite a broken plugin. Raise would stop xcube.
    warnings.warn(
        f"Unexpected exception while loading xcube plugin {entry_point.name!r}: {e}"
    )
    traceback.print_exc(file=sys.stderr)


class _ModuleEntryPoint:
    def __init__(self, module_name: str):
        self._module_name = module_name

    @property
    def name(self) -> str:
        return self._module_name

    def load(self) -> Optional[Callable]:
        """Load function "init_plugin()" either from "<module_name>.plugin" or "<module_name>"

        Returns:
            plugin init function
        """
        try:
            plugin_module = importlib.import_module(
                f"{self._module_name}.{PLUGIN_MODULE_NAME}"
            )
        except ModuleNotFoundError:
            plugin_module = importlib.import_module(self._module_name)

        module_func_name = PLUGIN_MODULE_FUNCTION_NAME
        if hasattr(plugin_module, module_func_name):
            module_func = getattr(plugin_module, module_func_name)
            if callable(module_func):
                return module_func

        # Here: The module name looks like an xcube-plugin
        # but lacks a callable named module_func_name

        return None


class ExtensionComponent(metaclass=abc.ABCMeta):
    """Utility base class for extension components."""

    def __init__(self, extension_point: str, name: str):
        if not extension_point:
            raise ValueError("extension_point must be given")
        if not name:
            raise ValueError("name must be given")
        self._extension_point = extension_point
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    def extension_point(self) -> str:
        """Returns:
        The extension point for this component.
        """
        return self._extension_point

    @property
    def extension(self) -> Optional[Extension]:
        """Returns:
        The extension for this component. None, if it has not (yet)
        been registered as an extension.
        """
        return get_extension_registry().get_extension(self._extension_point, self._name)

    def get_metadata_attr(self, key: str, default: Any = None) -> Any:
        """Returns:
        A metadata attribute for *key* and given *default* value.
        """
        extension = self.extension
        return extension.metadata.get(key, default) if extension is not None else ""



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Union

import numpy as np
import pyproj
import pyproj.transformer as pt
import xarray as xr

from .base import DEFAULT_TOLERANCE, GridMapping
from .coords import new_grid_mapping_from_coords
from .helpers import (
    Number,
    _assert_valid_xy_names,
    _normalize_crs,
    _normalize_number_pair,
)

# Cannot be used, but should, see TODO in transform_grid_mapping()
#
# class TransformedGridMapping(GridMapping, abc.ABC):
#     """
#     Grid mapping constructed from 1D/2D coordinate variables and a CRS.
#     """
#
#     def __init__(self,
#                  /,
#                  xy_coords: xr.DataArray,
#                  **kwargs):
#         self._xy_coords = xy_coords
#         super().__init__(**kwargs)
#
#     @property
#     def xy_coords(self) -> xr.DataArray:
#         return self._xy_coords
#


def transform_grid_mapping(
    grid_mapping: GridMapping,
    crs: Union[str, pyproj.crs.CRS],
    *,
    xy_res: Union[Number, tuple[Number, Number]] = None,
    tile_size: Union[int, tuple[int, int]] = None,
    xy_var_names: tuple[str, str] = None,
    tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    target_crs = _normalize_crs(crs)

    if xy_var_names:
        _assert_valid_xy_names(xy_var_names, name="xy_var_names")

    source_crs = grid_mapping.crs
    if source_crs == target_crs:
        if tile_size is not None or xy_var_names is not None:
            return grid_mapping.derive(tile_size=tile_size, xy_var_names=xy_var_names)
        return grid_mapping

    transformer = pt.Transformer.from_crs(source_crs, target_crs, always_xy=True)

    def _transform(block: np.ndarray) -> np.ndarray:
        x1, y1 = block
        x2, y2 = transformer.transform(x1, y1)
        return np.stack([x2, y2])

    xy_coords = xr.apply_ufunc(
        _transform,
        grid_mapping.xy_coords,
        output_dtypes=[np.float64],
        dask="parallelized",
    )
    if xy_res is not None:
        xy_bbox = transformer.transform_bounds(*grid_mapping.xy_bbox, densify_pts=101)
        x_res, y_res = _normalize_number_pair(xy_res)
        x_res_05, y_res_05 = x_res / 2, y_res / 2
        xy_bbox = (
            xy_bbox[0] - x_res_05,
            xy_bbox[1] - y_res_05,
            xy_bbox[2] + x_res_05,
            xy_bbox[3] + y_res_05,
        )
    else:
        xy_bbox = None

    xy_var_names = xy_var_names or ("transformed_x", "transformed_y")

    # TODO: Use a specialized grid mapping here that can store the
    #   *xy_coords* directly. Splitting the xy_coords dask array into
    #   x,y components as done here may be very inefficient for larger
    #   arrays, because x cannot be computed independently from y.
    #   This means, any access of x chunks will cause y chunks to be
    #   computed too and vice versa. As same operations are performed
    #   on x and y arrays, this will take twice as long as if operation
    #   would be performed on the xy_coords dask array directly.

    if tile_size is None:
        tile_size = grid_mapping.tile_size

    return new_grid_mapping_from_coords(
        x_coords=xy_coords[0].rename(xy_var_names[0]),
        y_coords=xy_coords[1].rename(xy_var_names[1]),
        crs=target_crs,
        xy_res=xy_res,
        xy_bbox=xy_bbox,
        tile_size=tile_size,
        tolerance=tolerance,
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import List, Optional

from xcube.util.assertions import assert_instance


class CubeGeneratorError(ValueError):
    """Represent a client or server error that
    may occur in the data cube generators.

    Args:
        args: arguments passed to base exceptions.
        status_code: Optional status code of the error (an integer). If
            given, HTTP error codes should be used. The range 400-499
            indicates client errors. The range 500-599 indicates server
            errors.
        remote_traceback: Traceback of an error occurred in a remote
            process.
        remote_output: Terminal output of a remote process.
    """

    def __init__(
        self,
        *args,
        status_code: Optional[int] = None,
        remote_traceback: Optional[list[str]] = None,
        remote_output: Optional[list[str]] = None,
        **kwargs,
    ):
        # noinspection PyArgumentList
        super().__init__(*args, **kwargs)
        if status_code is not None:
            assert_instance(status_code, int, "status_code")
        self._status_code = status_code
        self._remote_traceback = remote_traceback
        self._remote_output = remote_output

    @property
    def status_code(self) -> Optional[int]:
        """Status code of the error.
        May be None.
        """
        return self._status_code

    @property
    def remote_traceback(self) -> Optional[list[str]]:
        """Traceback of an error occurred in a remote process.
        May be None.
        """
        return self._remote_traceback

    @property
    def remote_output(self) -> Optional[list[str]]:
        """Terminal output of a remote process.
        May be None.
        """
        return self._remote_output



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Dict, Optional, Tuple, Union

from xcube.util.assertions import assert_instance, assert_true

# Make sure rfc3339-validator package is installed : jsonschema uses it for
# validating instances of JsonDateSchema and JsonDatetimeSchema.
# Use of __import__ avoids "unused package" warnings.
__import__("rfc3339_validator")

import jsonschema

from xcube.util.ipython import register_json_formatter
from xcube.util.undefined import UNDEFINED

Factory = Callable[..., Any]
Serializer = Callable[..., Any]

_TYPES_ENUM = {"null", "boolean", "integer", "number", "string", "array", "object"}
_NUMERIC_TYPES_ENUM = {"integer", "number"}


class JsonSchema(ABC):
    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        type: Union[str, Sequence[str]] = None,
        default: Any = UNDEFINED,
        const: Any = UNDEFINED,
        enum: Sequence[Any] = None,
        nullable: bool = None,
        title: str = None,
        description: str = None,
        examples: str = None,
        factory: Factory = None,
        serializer: Serializer = None,
    ):
        if type is not None and type not in _TYPES_ENUM:
            names = ", ".join(map(lambda t: f'"{t}"', sorted(list(_TYPES_ENUM))))
            raise ValueError(f"type must be one of {names}")
        if factory is not None and not callable(factory):
            raise ValueError("factory must be callable")
        if serializer is not None and not callable(serializer):
            raise ValueError("serializer must be callable")
        self.type: Optional[Union[str, Sequence[str]]] = type
        self.default: Any = default
        self.const: Any = const
        self.enum: Optional[Any] = list(enum) if enum is not None else None
        self.nullable: Optional[bool] = nullable
        self.title: Optional[str] = title
        self.description: Optional[str] = description
        self.examples: Optional[str] = examples
        self.factory = factory
        self.serializer = serializer

    def to_dict(self) -> dict[str, Any]:
        d = dict()
        if self.type is not None:
            if self.nullable is True and self.type != "null":
                d.update(type=[self.type, "null"])
            else:
                d.update(type=self.type)
        elif self.nullable is True:
            d.update(type="null")
        if self.default != UNDEFINED:
            d.update(default=self.default)
        if self.const != UNDEFINED:
            d.update(const=self.const)
        if self.enum is not None:
            d.update(enum=self.enum)
        if self.title is not None:
            d.update(title=self.title)
        if self.description is not None:
            d.update(description=self.description)
        if self.examples is not None:
            d.update(examples=self.examples)
        return d

    def validate_instance(self, instance: Any):
        """Validate JSON value *instance*."""
        # As a base for our custom validator, we hard-code a validator
        # version; JSON metaschema versions are not guaranteed to be backward
        # compatible, so if we use jsonschema.validators.validator_for(
        # schema), our schema may become invalid when the default validator
        # changes. The metaschema version can also be pinned in the schema
        # itself with an entry like this:
        # '$schema': 'http://json-schema.org/draft-07/schema'
        # However, this makes it more work to keep unit tests up to date, and
        # it's fiddly to adapt the recursive to_dict code to make sure that
        # the metaschema key *only* appears in the outermost dictionary.
        base_validator = jsonschema.validators.Draft7Validator

        # By default, jsonschema only recognizes lists as arrays. Here we derive
        # and use a custom validator which recognizes both lists and tuples as
        # arrays.
        new_type_checker = base_validator.TYPE_CHECKER.redefine(
            "array", lambda checker, inst: isinstance(inst, (list, tuple))
        )
        custom_validator = jsonschema.validators.extend(
            base_validator, type_checker=new_type_checker
        )

        # jsconschema needs extra packages installed to validate some formats;
        # if they are missing, the format check will be skipped silently. For
        # date-time format, strict_rfc3339 or rfc3339-validator is required.
        jsonschema.validate(
            instance=instance,
            schema=self.to_dict(),
            cls=custom_validator,
            format_checker=jsonschema.Draft7Validator.FORMAT_CHECKER,
        )

    def to_instance(self, value: Any) -> Any:
        """Convert Python object *value* into JSON value and return the validated result."""
        # TODO: support anyOf
        json_instance = self._to_unvalidated_instance(value)
        self.validate_instance(json_instance)
        return json_instance

    def from_instance(self, instance: Any) -> Any:
        """Validate JSON value *instance* and convert it into a Python object."""
        # TODO: support anyOf
        self.validate_instance(instance)
        return self._from_validated_instance(instance)

    @abstractmethod
    def _to_unvalidated_instance(self, value: Any) -> Any:
        """Turn Python object *value* into an unvalidated JSON value."""

    @abstractmethod
    def _from_validated_instance(self, instance: Any) -> Any:
        """Turn validated JSON value *instance* into a Python object."""


class JsonComplexSchema(JsonSchema):
    # TODO: implement JsonComplexTypeSchema more completely
    # For full support one_of, any_of, all_of should also be handled in from_instance and to_instance.

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        one_of: Sequence["JsonSchema"] = None,
        any_of: Sequence["JsonSchema"] = None,
        all_of: Sequence["JsonSchema"] = None,
        **kwargs,
    ):
        if len([x for x in (one_of, any_of, all_of) if bool(x)]) != 1:
            raise ValueError("exactly one of one_of, any_of, all_of must be given")
        super().__init__(**kwargs)
        self.one_of = one_of
        self.any_of = any_of
        self.all_of = all_of

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.one_of:
            d.update(oneOf=[schema.to_dict() for schema in self.one_of])
        if self.any_of:
            d.update(anyOf=[schema.to_dict() for schema in self.any_of])
        if self.all_of:
            d.update(allOf=[schema.to_dict() for schema in self.all_of])
        return d

    def _to_unvalidated_instance(self, value: Any) -> Any:
        return self.serializer(value) if self.serializer is not None else value

    def _from_validated_instance(self, instance: Any) -> Any:
        return self.factory(instance) if self.factory is not None else instance


class JsonSimpleSchema(JsonSchema):
    # noinspection PyShadowingBuiltins
    def __init__(self, type: str, **kwargs):
        super().__init__(type, **kwargs)
        if not isinstance(type, str):
            raise ValueError(f"illegal type: {type}")

    def _to_unvalidated_instance(self, value: Any) -> Any:
        return self.serializer(value) if self.serializer is not None else value

    def _from_validated_instance(self, instance: Any) -> Any:
        return self.factory(instance) if self.factory is not None else instance


class JsonNullSchema(JsonSimpleSchema):
    def __init__(self, **kwargs):
        super().__init__(type="null", **kwargs)


class JsonBooleanSchema(JsonSimpleSchema):
    def __init__(self, **kwargs):
        super().__init__(type="boolean", **kwargs)


class JsonStringSchema(JsonSimpleSchema):
    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        format: str = None,
        pattern: str = None,
        min_length: int = None,
        max_length: int = None,
        **kwargs,
    ):
        super().__init__(type="string", **kwargs)
        self.format = format
        self.pattern = pattern
        self.min_length = min_length
        self.max_length = max_length

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.format is not None:
            d.update(format=self.format)
        if self.pattern is not None:
            d.update(pattern=self.pattern)
        if self.min_length is not None:
            d.update(minLength=self.min_length)
        if self.max_length is not None:
            d.update(maxLength=self.max_length)
        return d


class JsonDateAndTimeSchemaBase:
    # noinspection PyShadowingBuiltins
    @classmethod
    def _validate_value(cls, value: Optional[str], name: str, format: str):
        if value is not None and not cls._is_valid_value(value, format):
            raise ValueError(f'{name} must be formatted as a "{format}"')

    # noinspection PyShadowingBuiltins
    @classmethod
    def _is_valid_value(cls, value: str, format: str) -> bool:
        try:
            jsonschema.validate(
                value,
                dict(type="string", format=format),
                format_checker=jsonschema.Draft7Validator.FORMAT_CHECKER,
            )
            return True
        except jsonschema.ValidationError:
            return False


class JsonDateSchema(JsonStringSchema, JsonDateAndTimeSchemaBase):
    """JSON schema for date instances.

    Args:
        min_date: optional minimum date.
        max_date: optional maximum date.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, min_date: str = None, max_date: str = None, **kwargs):
        super().__init__(**kwargs, format="date")
        self._validate_value(min_date, "min_date", "date")
        self._validate_value(max_date, "max_date", "date")
        self.min_date = min_date
        self.max_date = max_date

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.min_date is not None:
            d.update(minDate=self.min_date)
        if self.max_date is not None:
            d.update(maxDate=self.max_date)
        return d

    @classmethod
    def new_range(
        cls, min_date: str = None, max_date: str = None, nullable: bool = False
    ) -> "JsonArraySchema":
        """Return a schema for a date range.

        Args:
            min_date: optional minimum date.
            max_date: optional maximum date.
            nullable: whether the whole range as well as individual
                start and end may be None.

        Returns:
            a JsonArraySchema with two items.
        """
        return JsonArraySchema(
            items=[
                JsonDateSchema(min_date=min_date, max_date=max_date, nullable=nullable),
                JsonDateSchema(min_date=min_date, max_date=max_date, nullable=nullable),
            ],
            nullable=nullable,
        )


class JsonDatetimeSchema(JsonStringSchema, JsonDateAndTimeSchemaBase):
    """JSON schema for date-time instances.

    Args:
        min_datetime: optional minimum date-time.
        max_datetime: optional maximum date-time.
    """

    # noinspection PyShadowingBuiltins
    def __init__(self, min_datetime: str = None, max_datetime: str = None, **kwargs):
        super().__init__(**kwargs, format="date-time")
        self._validate_value(min_datetime, "min_datetime", "date-time")
        self._validate_value(max_datetime, "max_datetime", "date-time")
        self.min_datetime = min_datetime
        self.max_datetime = max_datetime

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.min_datetime is not None:
            d.update(minDatetime=self.min_datetime)
        if self.max_datetime is not None:
            d.update(maxDatetime=self.max_datetime)
        return d

    @classmethod
    def new_range(
        cls, min_datetime: str = None, max_datetime: str = None, nullable: bool = False
    ) -> "JsonArraySchema":
        """Return a schema for a date-time range.

        Args:
            min_datetime: optional minimum date.
            max_datetime: optional maximum date.
            nullable: whether the whole range as well as individual
                start and end dates may be None.

        Returns:
            a JsonArraySchema with two items.
        """
        return JsonArraySchema(
            items=[
                JsonDatetimeSchema(
                    min_datetime=min_datetime,
                    max_datetime=max_datetime,
                    nullable=nullable,
                ),
                JsonDatetimeSchema(
                    min_datetime=min_datetime,
                    max_datetime=max_datetime,
                    nullable=nullable,
                ),
            ],
            nullable=nullable,
        )


class JsonNumberSchema(JsonSimpleSchema):
    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        type: str = "number",
        minimum: Union[int, float] = None,
        exclusive_minimum: Union[int, float] = None,
        maximum: Union[int, float] = None,
        exclusive_maximum: Union[int, float] = None,
        multiple_of: Union[int, float] = None,
        **kwargs,
    ):
        if type not in _NUMERIC_TYPES_ENUM:
            names = ", ".join(sorted(map(lambda t: f'"{t}"', _NUMERIC_TYPES_ENUM)))
            raise ValueError(f"Type must be one of {names}")
        super().__init__(type, **kwargs)
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.multiple_of = multiple_of

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.minimum is not None:
            d.update(minimum=self.minimum)
        if self.exclusive_minimum is not None:
            d.update(exclusiveMinimum=self.exclusive_minimum)
        if self.maximum is not None:
            d.update(maximum=self.maximum)
        if self.exclusive_maximum is not None:
            d.update(exclusiveMaximum=self.exclusive_maximum)
        if self.multiple_of is not None:
            d.update(multipleOf=self.multiple_of)
        return d


class JsonIntegerSchema(JsonNumberSchema):
    def __init__(self, **kwargs):
        super().__init__(type="integer", **kwargs)


class JsonArraySchema(JsonSchema):
    def __init__(
        self,
        items: Union[JsonSchema, Sequence[JsonSchema]] = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = None,
        **kwargs,
    ):
        super().__init__(type="array", **kwargs)
        self.items = items
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.items is not None:
            if isinstance(self.items, JsonSchema):
                d.update(items=self.items.to_dict())
            else:
                d.update(items=[item.to_dict() for item in self.items])
        if self.min_items is not None:
            d.update(minItems=self.min_items)
        if self.max_items is not None:
            d.update(maxItems=self.max_items)
        if self.unique_items is not None:
            d.update(uniqueItems=self.unique_items)
        return d

    def _to_unvalidated_instance(
        self, value: Optional[Sequence[Any]]
    ) -> Optional[Sequence[Any]]:
        if self.serializer:
            return self.serializer(value)
        return self._convert_sequence(value, "_to_unvalidated_instance")

    def _from_validated_instance(self, instance: Sequence[Any]) -> Any:
        obj = self._convert_sequence(instance, "_from_validated_instance")
        return (
            self.factory(obj) if self.factory is not None and obj is not None else obj
        )

    def _convert_sequence(
        self, sequence: Optional[Sequence[Any]], method_name: str
    ) -> Optional[Sequence[Any]]:
        items_schema = self.items
        if sequence is None:
            # Sequence may be null too
            return None
        if isinstance(items_schema, JsonSchema):
            # Sequence turned into list with items_schema applying to all elements
            return [getattr(items_schema, method_name)(item) for item in sequence]
        if items_schema is not None:
            # Sequence turned into tuple with schema for every position
            return [
                getattr(items_schema[item_index], method_name)(sequence[item_index])
                for item_index in range(len(items_schema))
            ]
        # Sequence returned as-is, without schema
        return sequence


class JsonObjectSchema(JsonSchema):
    # TODO: also address property dependencies

    def __init__(
        self,
        properties: Mapping[str, JsonSchema] = None,
        additional_properties: Union[bool, JsonSchema] = None,
        min_properties: int = None,
        max_properties: int = None,
        required: Sequence[str] = None,
        dependencies: Mapping[str, Union[Sequence[str], JsonSchema]] = None,
        **kwargs,
    ):
        super().__init__(type="object", **kwargs)
        self.properties = dict(properties) if properties else dict()
        self.additional_properties = additional_properties
        self.min_properties = min_properties
        self.max_properties = max_properties
        self.required = list(required) if required else []
        self.dependencies = dict(dependencies) if dependencies else None

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        if self.properties is not None:
            d.update(properties={k: v.to_dict() for k, v in self.properties.items()})
        if self.additional_properties is not None:
            d.update(
                additionalProperties=(
                    self.additional_properties.to_dict()
                    if isinstance(self.additional_properties, JsonSchema)
                    else self.additional_properties
                )
            )
        if self.min_properties is not None:
            d.update(minProperties=self.min_properties)
        if self.max_properties is not None:
            d.update(maxProperties=self.max_properties)
        if self.required:
            d.update(required=list(self.required))
        if self.dependencies:
            d.update(
                dependencies={
                    k: (v.to_dict() if isinstance(v, JsonSchema) else v)
                    for k, v in self.dependencies.items()
                }
            )
        return d

    # TODO: move away. this is a special-purpose utility
    def process_kwargs_subset(
        self, kwargs: dict[str, Any], keywords: Sequence[str]
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Utility that helps to process keyword-arguments.

        Pops every keyword in *keywords* contained in this object schema's
        properties from *kwargs* and put the keyword and value from *kwargs*
        into a new dictionary.

        Return a 2-element tuple, first *kwargs* with the keywords
        from *keywords*., second *kwargs* without the keywords
        from *keywords*.

        The original *kwargs* is not touched.

        Returns: a tuple of two new keyword-argument dictionaries
        """
        old_kwargs = dict(kwargs)
        new_kwargs = {}
        for k in keywords:
            if k in old_kwargs:
                new_kwargs[k] = old_kwargs.pop(k)
            elif k in self.properties:
                property_schema = self.properties[k]
                if property_schema.const != UNDEFINED:
                    new_kwargs[k] = property_schema.const
                elif property_schema.default != UNDEFINED and k in self.required:
                    if property_schema.default != UNDEFINED:
                        new_kwargs[k] = property_schema.default
        return new_kwargs, old_kwargs

    def _to_unvalidated_instance(self, value: Any) -> Optional[Mapping[str, Any]]:
        if self.serializer is not None:
            return self.serializer(value)
        return self._convert_mapping(value, "_to_unvalidated_instance")

    def _from_validated_instance(self, instance: Optional[Mapping[str, Any]]) -> Any:
        obj = self._convert_mapping(instance, "_from_validated_instance")
        return (
            self.factory(**obj) if self.factory is not None and obj is not None else obj
        )

    def _convert_mapping(
        self, mapping_or_obj: Optional[Mapping[str, Any]], method_name: str
    ) -> Optional[Mapping[str, Any]]:
        # TODO: recognise self.dependencies. if dependency is again a schema, compute effective schema
        #       by merging this and the dependency.

        if mapping_or_obj is None:
            return None

        # Fix for #432: call object's own to_dict() method but beware of infinite recursion
        if (
            hasattr(mapping_or_obj, "to_dict")
            and callable(mapping_or_obj.to_dict)
            # calling JsonSchema.to_dict() is always fine
            and (
                isinstance(mapping_or_obj, JsonSchema)
                # calling JsonObject.to_dict() would cause infinite recursion
                or not isinstance(mapping_or_obj, JsonObject)
            )
        ):
            # noinspection PyBroadException
            try:
                return mapping_or_obj.to_dict()
            except BaseException:
                pass

        required_set = set(self.required) if self.required else set()

        additional_properties = self.additional_properties
        if additional_properties is None:
            # As of JSON Schema, additional_properties defaults to True
            additional_properties = True

        additional_properties_schema = None
        if isinstance(additional_properties, JsonSchema):
            additional_properties_schema = additional_properties

        converted_mapping = dict()

        if isinstance(mapping_or_obj, (collections.abc.Mapping, dict)):
            # An object that is already a Mapping
            mapping = mapping_or_obj
        else:
            # An object that is not a Mapping: convert to mapping.
            mapping = {}
            for property_name in dir(mapping_or_obj):
                property_value = UNDEFINED
                property_ok = False
                if property_name in self.properties:
                    # If property_name is defined in properties, we fully trust it
                    property_value = getattr(mapping_or_obj, property_name)
                    property_ok = True
                elif additional_properties and not property_name.startswith("_"):
                    # If property_name is not defined in properties, but additional
                    # properties are allowed, filter out private variables and callables.
                    property_value = getattr(mapping_or_obj, property_name)
                    property_ok = not callable(property_value)
                property_ok = property_ok and (
                    property_name in required_set or property_value is not None
                )
                if property_ok and property_value != UNDEFINED:
                    mapping[property_name] = property_value

        # recursively convert mapping into converted_mapping according to schema

        # process defined properties
        for property_name, property_schema in self.properties.items():
            if property_name in mapping:
                property_value = mapping[property_name]
                converted_property_value = getattr(property_schema, method_name)(
                    property_value
                )
                converted_mapping[property_name] = converted_property_value
            else:
                property_value = property_schema.default
                if property_value != UNDEFINED:
                    converted_property_value = getattr(property_schema, method_name)(
                        property_value
                    )
                    if (
                        property_name in required_set
                        or converted_property_value is not None
                    ):
                        converted_mapping[property_name] = converted_property_value

        # process additional properties
        if additional_properties:
            for property_name, property_value in mapping.items():
                if (
                    property_name not in converted_mapping
                    and property_value != UNDEFINED
                ):
                    if additional_properties_schema:
                        property_value = getattr(
                            additional_properties_schema, method_name
                        )(property_value)
                    converted_mapping[property_name] = property_value

        return converted_mapping

    @classmethod
    def inject_attrs(cls, obj: object, attrs: dict[str, Any]):
        for k, v in (attrs or {}).items():
            setattr(obj, k, v)


class JsonObject(ABC):
    """The abstract base class for objects

    * whose instances can be created from a JSON-serializable
      dictionary using their :meth:`from_dict` class method;
    * whose instances can be converted into a JSON-serializable dictionary
      using their :meth:`to_dict` instance method.

    Derived concrete classes must only implement the :meth:`get_schema` class method
    that must return a :class:`JsonObjectSchema`.

    Instances of this class have a JSON representation in Jupyter/IPython Notebooks.
    """

    @classmethod
    @abstractmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get JSON object schema."""

    @classmethod
    def from_dict(cls, value: dict[str, Any]) -> "JsonObject":
        """Create instance from JSON-serializable dictionary *value*."""
        return cls.get_schema().from_instance(value)

    def to_dict(self) -> dict[str, Any]:
        """Create JSON-serializable dictionary representation."""
        return self.get_schema().to_instance(self)

    def _inject_attrs(self, attrs: dict[str, Any]):
        assert_instance(attrs, dict, name="attrs")
        schema = self.get_schema()
        assert_true(
            isinstance(schema, JsonObjectSchema),
            message="schema must be a JSON object schema",
        )
        all_attrs = {k: None for k in (schema.properties or {}).keys()}
        all_attrs.update(attrs)
        JsonObjectSchema.inject_attrs(self, all_attrs)


register_json_formatter(JsonSchema)
register_json_formatter(JsonObject)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Mapping
from typing import Dict, Optional

from xcube.constants import LOG
from xcube.core.tile import (
    DEFAULT_CRS_NAME,
    DEFAULT_FORMAT,
    TileNotFoundException,
    TileRequestException,
    compute_rgba_tile,
)
from xcube.core.tilingscheme import DEFAULT_TILE_SIZE
from xcube.server.api import ApiError
from xcube.util.perf import measure_time_cm

from .context import TilesContext


def compute_ml_dataset_tile(
    ctx: TilesContext,
    ds_id: str,
    var_name: str,
    crs_name: Optional[str],
    x: str,
    y: str,
    z: str,
    params: Mapping[str, str],
):
    params = dict(params)
    trace_perf = params.pop("debug", "1" if ctx.datasets_ctx.trace_perf else "0") == "1"
    measure_time = measure_time_cm(logger=LOG, disabled=not trace_perf)
    with measure_time("Computing RGBA tile"):
        return _compute_ml_dataset_tile(
            ctx, ds_id, var_name, crs_name, x, y, z, params, trace_perf
        )


def _compute_ml_dataset_tile(
    ctx: TilesContext,
    ds_id: str,
    var_name: str,
    crs_name: Optional[str],
    x: str,
    y: str,
    z: str,
    args: dict[str, str],
    trace_perf: bool,
):
    try:
        x, y, z = (int(c) for c in (x, y, z))
    except ValueError:
        raise ApiError.BadRequest("x, y, z must be integers")

    crs_name = args.pop("crs", crs_name or DEFAULT_CRS_NAME)
    retina = args.pop("retina", None) == "1"
    cmap_name = args.pop("cmap", args.pop("cbar", None))
    cmap_norm = args.pop("norm", None)
    value_min = float(args.pop("vmin", 0.0))
    value_max = float(args.pop("vmax", 1.0))
    format = args.pop("format", DEFAULT_FORMAT)

    if format not in ("png", "image/png"):
        raise ApiError.BadRequest(f"Illegal format {format!r}")

    ml_dataset = ctx.datasets_ctx.get_ml_dataset(ds_id)
    if var_name == "rgb":
        var_names, value_ranges = ctx.datasets_ctx.get_rgb_color_mapping(
            ds_id, norm_range=(value_min, value_max)
        )
        components = "r", "g", "b"
        for i, c in enumerate(components):
            var_names[i] = args.pop(c, var_names[i])
            if var_names[i] is None:
                raise ApiError.BadRequest(
                    f"No variable in dataset {ds_id!r}"
                    f" specified for RGB component {c.upper()}"
                )
            value_ranges[i] = (
                float(args.pop(f"{c}vmin", value_ranges[i][0])),
                float(args.pop(f"{c}vmax", value_ranges[i][1])),
            )
    else:
        if (
            cmap_name is None
            or cmap_norm is None
            or value_min is None
            or value_max is None
        ):
            default_cmap_name, default_cmap_norm, default_value_range = (
                ctx.datasets_ctx.get_color_mapping(ds_id, var_name)
            )
            default_value_min, default_value_max = default_value_range
            if cmap_name is None:
                cmap_name = default_cmap_name
            if cmap_norm is None:
                cmap_norm = default_cmap_norm
            if value_min is None:
                value_min = default_value_min
            if value_max is None:
                value_max = default_value_max
        var_names = (var_name,)
        value_ranges = ((value_min, value_max),)

    try:
        return compute_rgba_tile(
            ml_dataset,
            var_names,
            x,
            y,
            z,
            ctx.datasets_ctx.colormap_registry,
            crs_name=crs_name,
            tile_size=(2 if retina else 1) * DEFAULT_TILE_SIZE,
            cmap_name=cmap_name,
            cmap_norm=cmap_norm,
            value_ranges=value_ranges,
            non_spatial_labels=args,
            format=format,
            trace_perf=trace_perf,
        )
    except TileNotFoundException as e:
        raise ApiError.NotFound(f"{e}") from e
    except TileRequestException as e:
        raise ApiError.BadRequest(f"{e}") from e



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import pkgutil
import sys
from string import Template

from xcube.server.api import ApiError, ApiHandler

from ...constants import LOG
from .api import api
from .context import MetaContext
from .controllers import get_service_info


@api.route("/")
class ServiceInfoHandler(ApiHandler[MetaContext]):
    @api.operation(
        operation_id="getServiceInfo", summary="Get information about the service"
    )
    def get(self):
        self.response.finish(get_service_info(self.ctx))


@api.route("/openapi.html")
class OpenApiHtmlHandler(ApiHandler):
    @api.operation(operation_id="getOpenApiHtml", summary="Show API documentation")
    def get(self):
        include_all = self.request.get_query_arg("all", default=False)
        html_template = pkgutil.get_data(
            "xcube.webapi.meta", "data/openapi.html"
        ).decode("utf-8")
        self.response.finish(
            Template(html_template).substitute(
                open_api_url=self.request.url_for_path(
                    "openapi.json", query="all=1" if include_all else None, reverse=True
                )
            )
        )


@api.route("/openapi.json")
class OpenApiJsonHandler(ApiHandler):
    @api.operation(
        operation_id="getOpenApiJson",
        summary="Get API documentation as OpenAPI 3.0 JSON document",
    )
    def get(self):
        include_all = self.request.get_query_arg("all", default=False)
        self.response.finish(
            self.ctx.get_open_api_doc(include_all=include_all),
            content_type="application/vnd.oai.openapi+json;version=3.0",
        )


@api.route("/maintenance/fail")
class MaintenanceFailHandler(ApiHandler[MetaContext]):
    @api.operation(
        operation_id="forceError",
        summary="Force a request error (for testing)",
        parameters=[
            {
                "name": "code",
                "in": "query",
                "description": "HTTP status code",
                "schema": {
                    "type": "integer",
                    "minimum": 400,
                },
            },
            {
                "name": "message",
                "in": "query",
                "description": "HTTP error message",
                "schema": {
                    "type": "string",
                },
            },
        ],
    )
    def get(self):
        """If *code* is given, the operation fails with
        that HTTP status code. Otherwise, the operation causes
        an internal server error.
        """
        code = self.request.get_query_arg("code", type=int)
        message = self.request.get_query_arg(
            "message", default="Error! No worries, this is just a test."
        )
        LOG.warning("Forcing error in request...")
        if code is None:
            raise RuntimeError(message)
        else:
            raise ApiError(code, message=message)


@api.route("/maintenance/update")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(
        operation_id="updateServer",
        summary="Force server update,"
        " updates configuration, and"
        " resets all resource caches.",
    )
    async def get(self):
        LOG.warning("Forcing server update...")
        # self.ctx.call_later(0,
        #                     self.ctx.server_ctx.server.update,
        #                     self.ctx.config)
        # self.ctx.server_ctx.server.update(self.ctx.config)
        await self.ctx.run_in_executor(
            None, self.ctx.server_ctx.server.update, self.ctx.config
        )
        return self.response.finish(dict(status="OK"))


@api.route("/maintenance/kill")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(
        operation_id="killServer", summary="Force server to shut down immediately."
    )
    def get(self):
        LOG.warning("Forcing server death...")
        sys.exit(0)


@api.route("/maintenance/oom")
class MaintenanceUpdateHandler(ApiHandler[MetaContext]):
    @api.operation(
        operation_id="forceOutOfMemory",
        summary="Force an out-of-memory error (for testing).",
    )
    def get(self):
        LOG.warning("Forcing out-of-memory error...")
        my_bytes = []
        from itertools import count

        for i in count(start=0, step=1):
            my_bytes.append(bytes(2**i))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


from typing import Dict, Union

import pyproj

CrsLike = Union[str, pyproj.CRS]


class ProjCache:
    """
    A cache for pyproj objects
    that may take considerable time to construct.
    """

    INSTANCE: "ProjCache"

    def __init__(self):
        self._crs_cache: dict[str, pyproj.CRS] = dict()
        self._transformer_cache: dict[str, pyproj.Transformer] = dict()

    def get_crs(self, crs: CrsLike) -> pyproj.CRS:
        if isinstance(crs, pyproj.CRS):
            return crs
        key = crs
        if key not in self._crs_cache:
            # pyproj.CRS.from_string() is expensive
            # save result for later use
            self._crs_cache[key] = pyproj.CRS.from_string(crs)
        return self._crs_cache[key]

    def get_transformer(self, crs1: CrsLike, crs2: CrsLike) -> pyproj.Transformer:
        crs1_key = self.get_crs_srs(crs1)
        crs2_key = self.get_crs_srs(crs2)
        key = f"{crs1_key}->{crs2_key}"
        if key not in self._transformer_cache:
            crs1 = self.get_crs(crs1)
            crs2 = self.get_crs(crs2)
            # pyproj.Transformer.from_crs() is really expensive (~0.5 secs)
            # save result for later use
            self._transformer_cache[key] = pyproj.Transformer.from_crs(
                crs1, crs2, always_xy=True
            )
        return self._transformer_cache[key]

    @classmethod
    def get_crs_srs(cls, crs: CrsLike) -> str:
        if isinstance(crs, pyproj.CRS):
            return crs.srs
        return crs


ProjCache.INSTANCE = ProjCache()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiError, ApiHandler

from .api import api
from .config import DEFAULT_COLLECTION_ID, PATH_PREFIX
from .context import StacContext
from .controllers import (
    get_collection,
    get_collection_item,
    get_collection_queryables,
    get_collection_schema,
    get_collections,
    get_conformance,
    get_datasets_collection_items,
    get_root,
    get_single_collection_items,
    search,
)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX, slash=True)
class CatalogRootHandler(ApiHandler[StacContext]):
    @api.operation(
        operation_id="getCatalogRoot", summary="Get the STAC catalog's root."
    )
    async def get(self):
        result = get_root(self.ctx.datasets_ctx, self.request.reverse_base_url)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/conformance", slash=True)
class CatalogConformanceHandler(ApiHandler[StacContext]):
    @api.operation(
        operation_id="getCatalogConformance",
        summary="Get the STAC catalog's conformance.",
    )
    async def get(self):
        result = get_conformance()
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections", slash=True)
class CatalogCollectionsHandler(ApiHandler[StacContext]):
    @api.operation(
        operation_id="getCatalogCollections",
        summary="Get the STAC catalog's collections.",
    )
    async def get(self):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        result = get_collections(
            self.ctx, self.request.reverse_base_url, granted_scopes=granted_scopes
        )
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}", slash=True)
class CatalogCollectionHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getCatalogCollection", summary="Get a STAC catalog collection."
    )
    async def get(self, collectionId: str):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        result = get_collection(
            self.ctx,
            self.request.reverse_base_url,
            collectionId,
            granted_scopes=granted_scopes,
        )
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/items", slash=True)
class CatalogCollectionItemsHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getCatalogCollectionItems",
        summary="Get the items of a STAC catalog collection.",
    )
    async def get(self, collectionId: str):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        get_items_args = dict(
            ctx=self.ctx.datasets_ctx,
            base_url=self.request.reverse_base_url,
            collection_id=collectionId,
            granted_scopes=granted_scopes,
        )
        if collectionId == DEFAULT_COLLECTION_ID:
            if "limit" in self.request.query:
                get_items_args["limit"] = self.request.get_query_arg("limit", type=int)
            if "cursor" in self.request.query:
                get_items_args["cursor"] = self.request.get_query_arg(
                    "cursor", type=int
                )
            result = get_datasets_collection_items(**get_items_args)
        else:
            result = get_single_collection_items(**get_items_args)

        return await self.response.finish(result, content_type="application/geo+json")


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/items/{featureId}", slash=True)
class CatalogCollectionItemHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="getCatalogCollectionItem",
        summary="Get an item of a STAC catalog collection.",
    )
    async def get(self, collectionId: str, featureId: str):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(self.request.headers)
        result = get_collection_item(
            self.ctx.datasets_ctx,
            self.request.reverse_base_url,
            collectionId,
            featureId,
            granted_scopes=granted_scopes,
        )
        return await self.response.finish(result, content_type="application/geo+json")


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/search", slash=True)
class CatalogSearchHandler(ApiHandler[StacContext]):
    @api.operation(
        operation_id="searchCatalogByKeywords",
        summary="Search the STAC catalog by keywords.",
    )
    async def get(self):
        # TODO (forman): get search params from query
        result = search(self.ctx.datasets_ctx, self.request.reverse_base_url)
        return await self.response.finish(result)

    @api.operation(
        operation_id="searchCatalogByJSON",
        summary="Search the STAC catalog by JSON request body.",
    )
    async def post(self):
        # TODO (forman): get search params from request body
        result = search(self.ctx.datasets_ctx, self.request.reverse_base_url)
        return await self.response.finish(result)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/queryables", slash=True)
class QueryablesHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="queryables",
        summary="Return a JSON Schema defining the supported "
        'metadata filters (also called "queryables") '
        "for a specific collection.",
    )
    async def get(self, collectionId):
        schema = get_collection_queryables(
            ctx=self.ctx.datasets_ctx, collection_id=collectionId
        )
        return await self.response.finish(schema)


# noinspection PyAbstractClass,PyMethodMayBeStatic
@api.route(PATH_PREFIX + "/collections/{collectionId}/schema", slash=True)
class SchemaHandler(ApiHandler[StacContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operation_id="schema",
        summary="Return a JSON Schema defining the range "
        "(i.e. data variables) of a specific collection.",
    )
    async def get(self, collectionId):
        schema = get_collection_schema(
            ctx=self.ctx.datasets_ctx,
            base_url=self.request.reverse_base_url,
            collection_id=collectionId,
        )
        return await self.response.finish(schema)
